import numpy as np
from astropy.time.core import Time
from datetime import datetime
import h5py
import os

from pyspi.io.package_data import get_path_of_internal_data_dir
from pyspi.utils.function_utils import get_time_object

double_names = {19: [0, 1], 20: [0, 2], 21: [0, 3], 22: [0, 4], 23: [0, 5],
                24: [0, 6], 25: [1, 2], 26: [1, 6], 27: [1, 7], 28: [1, 8],
                29: [1, 9], 30: [2, 3], 31: [2, 9], 32: [2, 10], 33: [2, 11],
                34: [3, 4], 35: [3, 11], 36: [3, 12], 37: [3, 13], 38: [4, 5],
                39: [4, 13], 40: [4, 14], 41: [4, 15], 42: [5, 6], 43: [5, 15],
                44: [5, 16], 45: [5, 17], 46: [6, 7], 47: [6, 17], 48: [6, 18],
                49: [7, 8], 50: [7, 18], 51: [8, 9], 52: [9, 10], 53: [10, 11],
                54: [11, 12], 55: [12, 13], 56: [13, 14], 57: [14, 15],
                58: [15, 16], 59: [16, 17], 60: [17, 18]
                }

triple_names = {61: [0, 1, 2], 62: [0, 2, 3], 63: [0, 3, 4], 64: [0, 4, 5],
                65: [0, 5, 6], 66: [0, 6, 1], 67: [1, 2, 9],
                68: [1, 6, 7], 69: [1, 7, 8], 70: [1, 8, 9], 71: [2, 3, 11],
                72: [2, 9, 10], 73: [2, 10, 11], 74: [3, 4, 13],
                75: [3, 11, 12], 76: [3, 12, 13],
                77: [4, 5, 15], 78: [4, 13, 14],
                79: [4, 14, 15], 80: [5, 6, 17],
                81: [5, 15, 16], 82: [5, 16, 17],
                83: [6, 7, 18], 84: [6, 17, 18]}


def get_live_dets(time, event_types=["single", "double", "triple"]):
    """
    Get the live dets for a given time

    :param time: Live dets at a given time. Either
        "YYMMDD HHMMSS" or as astropy time object
    :param event_types: which event types?
        List with single, double and/or triple

    :returns: array of live dets
    """

    time = get_time_object(time)

    # All single dets
    live_dets = np.arange(19)
    dead_dets = []
    # Check if time is after the failure times
    # (from https://www.isdc.unige.ch/integral/download/osa/doc/10.1/
    # osa_um_spi/node69.html )
    if time > Time(datetime.strptime('031206 060000', '%y%m%d %H%M%S')):
        live_dets = live_dets[live_dets != 2]
        dead_dets.append(2)
    if time > Time(datetime.strptime('040717 082006', '%y%m%d %H%M%S')):
        live_dets = live_dets[live_dets != 17]
        dead_dets.append(17)
    if time > Time(datetime.strptime('090219 095957', '%y%m%d %H%M%S')):
        live_dets = live_dets[live_dets != 5]
        dead_dets.append(5)
    if time > Time(datetime.strptime('100527 124500', '%y%m%d %H%M%S')):
        live_dets = live_dets[live_dets != 1]
        dead_dets.append(1)

    all_dets = np.array([])
    if "single" in event_types:
        all_dets = np.concatenate([all_dets, live_dets])

    if "double" in event_types:
        live_double_dets = []
        for key, value in zip(double_names.keys(),
                              double_names.values()):
            dead = False
            for v in value:
                if v in dead_dets:
                    dead = True

            if not dead:
                live_double_dets.append(key)
        all_dets = np.concatenate([all_dets,
                                   live_double_dets])
    if "triple" in event_types:
        live_triple_dets = []
        for key, value in zip(triple_names.keys(),
                              triple_names.values()):
            dead = False
            for v in value:
                if v in dead_dets:
                    dead = True

            if not dead:
                live_triple_dets.append(key)
        all_dets = np.concatenate([all_dets,
                                   live_triple_dets])
    return np.array(all_dets, dtype=int)


def get_live_dets_pointing(pointing,
                           event_types=["single", "double", "triple"]):
    """
    Get livedets for a given pointing id

    :param pointing: pointing id
    :param event_types: which event types?
        List with single, double and/or triple

    :returns:
    """
    # get end time of pointing
    id_file_path = os.path.join(get_path_of_internal_data_dir,
                                'id_data_time.hdf5')
    with h5py.File(id_file_path, "r") as f:
        idx = np.argwhere(f["ID"][()] == pointing.encode('utf-8'))
        assert len(idx) != 0, "Poiinting not found in database"
        idx = idx[0,0]

        isdc_mjd_time = f["Start"][idx]

    time = Time(isdc_mjd_time+51544, format='mjd', scale='utc')

    return get_live_dets(time, event_types)



# Test spi pointing

def test_spipointing():
    from pyspi.utils.response.spi_pointing import SPIPointing
    from pyspi.io.package_data import get_path_of_external_data_dir
    from pyspi.io.get_files import get_files
    import os
    import numpy as np
    
    pointing_id = 169600130010
    get_files(pointing_id)

    geom_save_path = os.path.join(get_path_of_external_data_dir(),
                                  'pointing_data',
                                  str(pointing_id),
                                  'sc_orbit_param.fits.gz')

    point = SPIPointing(geom_save_path)

    assert np.sum(point.sc_matrix) == 1222.1378651210798, \
        'SPI pointing test failed'



from astropy.time.core import Time, TimeDelta
import yaml
import os
import numpy as np
from datetime import datetime
import h5py
from astropy.io import fits


from pyspi.io.get_files import get_files
from pyspi.io.package_data import get_path_of_external_data_dir
from pyspi.utils.livedets import double_names, triple_names
from pyspi.utils.function_utils import find_needed_ids, ISDC_MJD_to_cxcsec, \
    get_time_object

from threeML.io.file_utils import sanitize_filename
from threeML.utils.time_series.event_list import EventListWithDeadTime,\
    EventListWithLiveTime
from threeML.utils.time_series.binned_spectrum_series import \
    BinnedSpectrumSeries
from threeML.utils.spectrum.binned_spectrum import BinnedSpectrumWithDispersion
from threeML.utils.data_builders.time_series_builder import TimeSeriesBuilder


class SPISWFile(object):
    def __init__(self, det, pointing_id, ebounds):
        """
        Class to read in all the data needed from a SCW file
        for a given config file

        :param config: Config yml filename, Config object or dict
        :param det: For which detector?

        """
        # General nameing
        self._det_name = f"Detector {det}"
        self._mission = "Integral/SPI"
        self._ebounds = ebounds
        # How many echans?
        self._n_channels = len(self._ebounds)-1

        # Check that det is a valid number
        self._det = det
        assert self._det in np.arange(85), f"{self._det} is not a valid"\
            " detector. Please only use detector ids between 0 and 84."

        # Find the SW ID of the data file we need for this time
        self._pointing_id = pointing_id

        # Get the data, either from afs or from ISDC archive
        try:
            # Get the data from the afs server
            get_files(pointing_id, access="afs")
        except AssertionError:
            # Get the files from the iSDC data archive
            print("AFS data access did not work."
                  " I will try the ISDC data archive.")
            get_files(pointing_id, access="isdc")

        # Read in all we need
        self._read_in_pointing_data(self._pointing_id)

    def _read_in_pointing_data(self, pointing_id):
        """
        Gets all needed information from the data file for the given
        pointing_id

        :param pointing_id: pointing_id for which we want the data

        :returns:
        """

        with fits.open(os.path.join(get_path_of_external_data_dir(),
                                    'pointing_data', pointing_id,
                                    'spi_oper.fits.gz')) as hdu_oper:

            # Get time of first and last event (t0 at grb time)
            time_sgl = ISDC_MJD_to_cxcsec(hdu_oper[1].data['time'])
            time_psd = ISDC_MJD_to_cxcsec(hdu_oper[2].data['time'])
            time_me2 = ISDC_MJD_to_cxcsec(hdu_oper[4].data['time'])
            time_me3 = ISDC_MJD_to_cxcsec(hdu_oper[5].data['time'])

            self._time_start = np.min(np.concatenate([time_sgl, time_psd,
                                                      time_me2, time_me3]))
            self._time_stop = np.max(np.concatenate([time_sgl, time_psd,
                                                     time_me2, time_me3]))

            # Read in the data for the wanted detector
            # For single events we have to take both the non_psd
            # (often called sgl here...)
            # and the psd events. Both added together
            # give the real single events.
            if self._det in range(19) or self._det == "singles":
                dets_sgl = hdu_oper[1].data['DETE']
                if self._det != "singles":
                    time_sgl = time_sgl[dets_sgl == self._det]
                    energy_sgl = hdu_oper[1].data['energy'][dets_sgl ==
                                                            self._det]
                else:
                    energy_sgl = hdu_oper[1].data['energy']

                dets_psd = hdu_oper[2].data['DETE']
                if self._det != "singles":
                    time_psd = time_psd[dets_psd == self._det]
                    energy_psd = hdu_oper[2].data['energy'][dets_psd ==
                                                            self._det]
                else:
                    energy_psd = hdu_oper[2].data['energy']

            if self._det in range(19, 61):
                dets_me2 = np.sort(hdu_oper[4].data['DETE'], axis=1)
                i, k = double_names[self._det]
                mask = np.logical_and(dets_me2[:, 0] == i,
                                      dets_me2[:, 1] == k)

                time_me2 = time_me2[mask]
                energy_me2 = np.sum(hdu_oper[4].data['energy'][mask], axis=1)

            if self._det in range(61,85):
                dets_me3 = np.sort(hdu_oper[5].data['DETE'], axis=1)
                i, j, k = triple_names[self._det]
                mask = np.logical_and(np.logical_and(dets_me3[:, 0] == i,
                                                     dets_me3[:, 1] == j),
                                      dets_me3[:, 2] == k)

                time_me3 = time_me3[mask]
                energy_me3 = np.sum(hdu_oper[5].data['energy'][mask], axis=1)

        if self._det in range(19):

            self._times = time_psd
            self._energies = energy_psd

            self._times = np.append(self._times, time_sgl)
            self._energies = np.append(self._energies, energy_sgl)
            # sort in time
            sort_array = np.argsort(self._times)
            self._times = self._times[sort_array]
            self._energies = self._energies[sort_array]

        if self._det in range(19, 61):

            self._times = time_me2
            self._energies = energy_me2

        if self._det in range(61, 85):

            self._times = time_me3
            self._energies = energy_me3

        # Check if there are any counts
        if np.sum(self._energies) == 0:

            raise AssertionError(f"The detector {self._det} has zero counts"
                                 " and is therefore not active."
                                 " Please exclude this detector!")

        # Bin this in the energy bins we have
        self._energy_bins = np.ones_like(self._energies, dtype=int)*-1
        # Loop over ebins
        for i, (emin, emax) in enumerate(zip(self._ebounds[:-1],
                                             self._ebounds[1:])):
            mask = np.logical_and(self._energies > emin,
                                  self._energies < emax)
            self._energy_bins[mask] = np.ones_like(self._energy_bins[mask])*i

        # Throw away all events that have energies outside of the ebounds that
        # should be used
        mask = self._energy_bins == -1
        self._energy_bins = self._energy_bins[~mask]
        self._times = self._times[~mask]
        self._energies = self._energies[~mask]

    @property
    def geometry_file_path(self):
        """
        :returns: Path to the spacecraft geometry file
        """
        return os.path.join(get_path_of_external_data_dir(), 'pointing_data',
                            self._pointing_id, 'sc_orbit_param.fits.gz')

    @property
    def times(self):
        """
        :returns: times of detected events
        """
        return self._times

    @property
    def energies(self):
        """
        :returns: energies of detected events
        """
        return self._energies

    @property
    def energy_bins(self):
        """
        :returns: energy bin number of every event
        """
        return self._energy_bins

    @property
    def ebounds(self):
        """
        :returns: ebounds of analysis
        """
        return self._ebounds

    @property
    def det(self):
        """
        :returns: detector ID
        """
        return self._det

    @property
    def n_channels(self):
        """
        :returns: number energy channels
        """
        return self._n_channels

    @property
    def time_start(self):
        """
        :returns: start time of lightcurve
        """
        return self._time_start

    @property
    def time_stop(self):
        """
        :returns: stop time of lightcurve
        """
        return self._time_stop

    @property
    def det_name(self):
        """
        :returns: Name det
        """
        return self._det_name

    @property
    def mission(self):
        """
        :returns: Name Mission
        """
        return self._mission


class SPISWFileGRB(object):

    def __init__(self, det, ebounds, time_of_grb, sgl_type=None):
        """
        Class to read in all the data needed from a SCW file for a given
        grbtime

        :param det: For which detector?
        :param ebounds: Ebounds for the Analysis.
        :param time_of_grb: Time of the GRB as "YYMMDD HHMMSS"
        :param sgl_type: Which type of single events?
        Only normal sgl, psd or both?

        :returns: Object
        """

        self._det_name = f"Detector {det}"
        self._mission = "Integral/SPI"

        # Set ebounds of energy bins
        self._ebounds = ebounds

        # How many echans?
        self._n_channels = len(self._ebounds)-1

        # Time of GRB. Needed to get the correct pointing.
        self._time_of_GRB = get_time_object(time_of_grb)

        # Check that det is a valid number
        self._det = det
        assert self._det in np.arange(85), f"{self._det} is not a valid"\
            "detector. Please only use detector ids between 0 and 84."

        if self._det < 19:
            assert sgl_type is not None, "Only PSD Events?"
            assert sgl_type in ["psd", "sgl", "both"], \
                "sgl_type must be psd, sgl or both"
            self._sgl_type = sgl_type
        # Find the SW ID of the data file we need for this time
        self._pointing_id = find_needed_ids(self._time_of_GRB)

        # Get the data, either from afs or from ISDC archive
        try:
            # Get the data from the afs server
            get_files(self._pointing_id, access="afs")
        except AssertionError:
            # Get the files from the iSDC data archive
            print("AFS data access did not work."
                  " I will try the ISDC data archive.")
            get_files(self._pointing_id, access="isdc")

        # Reference time of GRB
        self._GRB_ref_time_cxcsec =\
            ISDC_MJD_to_cxcsec((self._time_of_GRB).tt.mjd-51544)
        # Read in all we need
        self._read_in_pointing_data()
        self._get_deadtime_info()

    def _read_in_pointing_data(self):
        """
        Gets all needed information from the data file for the given
        pointing_id

        :param pointing_id: pointing_id for which we want the data

        :returns:
        """

        with fits.open(os.path.join(get_path_of_external_data_dir(),
                                    'pointing_data', self._pointing_id,
                                    'spi_oper.fits.gz')) as hdu_oper:

            # Get time of first and last event (t0 at grb time)
            time_sgl = ISDC_MJD_to_cxcsec(hdu_oper[1].data['time']) \
                - self._GRB_ref_time_cxcsec
            time_psd = ISDC_MJD_to_cxcsec(hdu_oper[2].data['time']) \
                - self._GRB_ref_time_cxcsec
            time_me2 = ISDC_MJD_to_cxcsec(hdu_oper[4].data['time']) \
                - self._GRB_ref_time_cxcsec
            time_me3 = ISDC_MJD_to_cxcsec(hdu_oper[5].data['time']) \
                - self._GRB_ref_time_cxcsec

            # Get time of first and last detected photon
            self._time_start = np.min(np.concatenate([time_sgl, time_psd,
                                                      time_me2, time_me3]))
            self._time_stop = np.max(np.concatenate([time_sgl, time_psd,
                                                     time_me2, time_me3]))

            # Read in the data for the wanted detector
            # For single events we have to take both the non_psd
            # (often called sgl here...) and the psd events.
            # Both added together give the real single events.
            if self._det in range(19):
                dets_sgl = hdu_oper[1].data['DETE']
                time_sgl = time_sgl[dets_sgl == self._det]
                energy_sgl = hdu_oper[1].data['energy'][dets_sgl == self._det]

                dets_psd = hdu_oper[2].data['DETE']
                time_psd = time_psd[dets_psd == self._det]
                energy_psd = hdu_oper[2].data['energy'][dets_psd == self._det]

            if self._det in range(19, 61):
                dets_me2 = np.sort(hdu_oper[4].data['DETE'], axis=1)
                i, k = double_names[self._det]
                mask = np.logical_and(dets_me2[:, 0] == i,
                                      dets_me2[:, 1] == k)

                time_me2 = time_me2[mask]
                energy_me2 = np.sum(hdu_oper[4].data['energy'][mask], axis=1)

            if self._det in range(61,85):
                dets_me3 = np.sort(hdu_oper[5].data['DETE'], axis=1)
                i, j, k = triple_names[self._det]
                mask = np.logical_and(np.logical_and(dets_me3[:, 0] == i,
                                                     dets_me3[:, 1] == j),
                                      dets_me3[:, 2] == k)

                time_me3 = time_me3[mask]
                energy_me3 = np.sum(hdu_oper[5].data['energy'][mask], axis=1)

        if self._det in range(19):

            self._times = np.array([])
            self._energies = np.array([])

            # Use the single events with the given flag
            if self._sgl_type == "psd" or self._sgl_type == "both":
                self._times = np.append(self._times, time_psd)
                self._energies = np.append(self._energies, energy_psd)

            if self._sgl_type == "sgl" or self._sgl_type == "both":
                self._times = np.append(self._times, time_sgl)
                self._energies = np.append(self._energies, energy_sgl)

            # sort in time
            sort_array = np.argsort(self._times)
            self._times = self._times[sort_array]
            self._energies = self._energies[sort_array]

        if self._det in range(19, 61):

            self._times = time_me2
            self._energies = energy_me2

        if self._det in range(61, 85):

            self._times = time_me3
            self._energies = energy_me3

        # Check if there are any counts
        if np.sum(self._energies) == 0:

            raise AssertionError(f"The detector {self._det} has zero counts "
                                 "and is therefore not active. "
                                 "Please exclude this detector!")

        # Assign the corresponsing energy channel number to every detected
        # energy
        self._energy_bins = np.ones_like(self._energies, dtype=int)*-1
        # Loop over ebins
        for i, (emin, emax) in enumerate(zip(self._ebounds[:-1],
                                             self._ebounds[1:])):
            mask = np.logical_and(self._energies > emin,
                                  self._energies < emax)
            self._energy_bins[mask] = np.ones_like(self._energy_bins[mask])*i

        # Throw away all events that have energies outside of the ebounds that
        # should be used
        mask = self._energy_bins == -1
        self._energy_bins = self._energy_bins[~mask]
        self._times = self._times[~mask]
        self._energies = self._energies[~mask]

    def _get_deadtime_info(self):
        """
        Get the deadtime info from the hk file

        :returns:
        """

        # read in the OB-Time (on-board time) and the ISDC-MJD times
        # of the events from the data file. We will need this to
        # convert the OB-Times given in the hk file into ISDC-MJD time
        with fits.open(os.path.join(get_path_of_external_data_dir(),
                                    'pointing_data',
                                    self._pointing_id,
                                    'spi_oper.fits.gz')) as hdu_oper:

            ob_time = hdu_oper["SPI.-OSGL-ALL"].data["OB_TIME"]
            time = hdu_oper["SPI.-OSGL-ALL"].data["TIME"]

        # calulate the true on-board time from the 4 given values
        # in the fits files. See:
        # (https://www.isdc.unige.ch/integral/support/faq.cgi?DATA-006)

        ob_time_data = (ob_time[:, 0]*655363**3 +
                        ob_time[:, 1]*655362**2 +
                        ob_time[:, 2]*65536 +
                        ob_time[:, 3])

        # read in hk file with the deadtime per 1 second intervall
        # in units of 100 nano seconds
        deadtimes = np.array([])
        with fits.open(os.path.join(get_path_of_external_data_dir(),
                                    'pointing_data',
                                    self._pointing_id,
                                    'spi_science_hk.fits.gz')) as hdu_oper:

            for i in range(19):
                deadtimes = np.append(deadtimes,
                                      hdu_oper["SPI.-SCHK-HRW"].
                                      data[f"P__DF__CAFDT__L{i}"])

            ob_time = hdu_oper["SPI.-SCHK-CNV"].data["OB_TIME"]

        # reshape to get shape (num_det, num_timebins)
        # and multipy by 100*10**-9 to get the deadtime in
        # seconds
        deadtimes = deadtimes.reshape(19, -1)*100*10**-9

        # again get true OB-Time
        ob_time_hk = (ob_time[:, 0]*655363**3 +
                      ob_time[:, 1]*655362**2 +
                      ob_time[:, 2]*65536 +
                      ob_time[:, 3])

        # use linear interpolation of OB-Time/ISDC-MJD pairs from
        # the data file to get the ISDC-MJD time of the hk time bins
        times_hk = np.interp(ob_time_hk, ob_time_data, time)

        # add end of last time bin (+1 second in units of days)
        times_hk = np.append(times_hk, times_hk[-1]+(1.0/(24*3600)))

        # save time and deadtime - add end of
        self._deadtime_bin_edges = ISDC_MJD_to_cxcsec(times_hk) - \
            self._GRB_ref_time_cxcsec
        self._deadtimes = deadtimes[self._det]


    @property
    def geometry_file_path(self):
        """
        :returns: Path to the spacecraft geometry file
        """
        return os.path.join(get_path_of_external_data_dir(),
                            'pointing_data',
                            self._pointing_id,
                            'sc_orbit_param.fits.gz')

    @property
    def times(self):
        """
        :returns: times of detected events
        """
        return self._times

    @property
    def energies(self):
        """
        :returns: energies of detected events
        """
        return self._energies

    @property
    def energy_bins(self):
        """
        :returns: energy bin number of every event
        """
        return self._energy_bins

    @property
    def ebounds(self):
        """
        :returns: ebounds of analysis
        """
        return self._ebounds

    @property
    def det(self):
        """
        :returns: detector ID
        """
        return self._det

    @property
    def n_channels(self):
        """
        :returns: number energy channels
        """
        return self._n_channels

    @property
    def time_start(self):
        """
        :returns: start time of lightcurve
        """
        return self._time_start

    @property
    def time_stop(self):
        """
        :returns: stop time of lightcurve
        """
        return self._time_stop

    @property
    def det_name(self):
        """
        :returns: Name det
        """
        return self._det_name

    @property
    def mission(self):
        """
        :returns: Name Mission
        """
        return self._mission

    @property
    def deadtime_bin_starts(self):
        """
        :returns: Start time of time bins which have the deadtime
        information
        """
        return self._deadtime_bin_edges[:-1]

    @property
    def deadtime_bin_stops(self):
        """
        :returns: Stop time of time bins which have the deadtime
        information
        """
        return self._deadtime_bin_edges[1:]

    @property
    def deadtimes_per_interval(self):
        """
        :returns: Deadtime per time bin which have the deadtime
        information
        """
        return self._deadtimes

    @property
    def livetimes_per_interval(self):
        """
        :returns: Livetime per time bin which have the deadtime
        information
        """
        return 1-self._deadtimes


class TimeSeriesBuilderSPI(TimeSeriesBuilder):

    def __init__(
        self,
        name,
        time_series,
        response=None,
        poly_order=-1,
        verbose=True,
        restore_poly_fit=None,
        container_type=BinnedSpectrumWithDispersion,
        **kwargs
    ):
        """
        Class to build the time_series for SPI. Inherited from the 3ML
        TimeSeriesBuilder with added class methods to build the object
        for SPI datafiles.
        :param name: Name of the tsb
        :param time_series: Timeseries with the data
        :param response: Response object
        :param poly_order: poly order for the polynominal fitting
        :param verbose: Verbose?
        :param restore_poly_fit: Path to a file with the poly bkg fits
        :param containter_type: ContainerType for spectrum
        :returns: Object
        """

        super(TimeSeriesBuilderSPI, self).__init__(
            name,
            time_series,
            response=response,
            poly_order=poly_order,
            unbinned=False,
            verbose=verbose,
            restore_poly_fit=restore_poly_fit,
            container_type=container_type,
            **kwargs
        )

    @classmethod
    def from_spi_grb(cls,
                     name,
                     det,
                     time_of_grb,
                     ebounds=None,
                     response=None,
                     sgl_type=None,
                     restore_background=None,
                     poly_order=0,
                     verbose=True
    ):
        """
        Class method to build the time_series_builder for a given GRB time

        :param name: Name of object
        :param det: Which det?
        :param ebounds: Output ebounds for analysis.
        :param time_of_grb: Astropy time object with the time of the GRB (t0)
        :param response: InstrumenResponse Object
        :param sgl_type: What kind of sinlge events? Standard single events?
                         PSD events? Or both?
        :param restore_background: File to restore bkg
        :param poly_order: Which poly_order? -1 gives automatic determination
        :param verbose: Verbose?

        :returns: Initalized TimeSeriesBuilder object
        """

        assert ebounds is not None or response is not None, "You have to "\
            "either specify ebounds or input a response object."

        if ebounds is not None and response is not None:
            # check that the ebounds match
            assert np.all(response.ebounds == ebounds), "Ebounds do not "\
                "match the ones in the response object!"

        elif response is not None:
            # ebounds is None in this case
            ebounds = response.ebounds

        # Get an object with all the needed information
        spi_grb_setup = SPISWFileGRB(det,
                                     ebounds,
                                     time_of_grb,
                                     sgl_type=sgl_type)

        event_list = EventListWithLiveTime(
            arrival_times=spi_grb_setup.times,
            measurement=spi_grb_setup.energy_bins,
            n_channels=spi_grb_setup.n_channels,
            start_time=spi_grb_setup.time_start,
            stop_time=spi_grb_setup.time_stop,
            live_time=spi_grb_setup.livetimes_per_interval,
            live_time_starts=spi_grb_setup.deadtime_bin_starts,
            live_time_stops=spi_grb_setup.deadtime_bin_stops,
            first_channel=0,
            instrument=spi_grb_setup.det_name,
            mission=spi_grb_setup.mission,
            verbose=verbose,
            edges=spi_grb_setup.ebounds,
        )

        return cls(
            name,
            event_list,
            poly_order=poly_order,
            verbose=verbose,
            restore_poly_fit=restore_background,
            response=response,
            container_type=BinnedSpectrumWithDispersion
        )

    @classmethod
    def from_spi_constant_pointing(cls,
                                   det,
                                   ebounds,
                                   pointing_id,
                                   response
    ):
        """
        Class method to build the time_series_builder for a given pointing id

        :param det: Which det?
        :param ebounds: Output ebounds for analysis.
        :param pointing_id: Pointing ID
        :param response: InstrumenResponse Object

        :returns: Initalized TimeSeriesBuilder object
        """

        spi_grb_setup1 = SPISWFile(det, pointing_id, ebounds)

        e = EventListWithDeadTime(arrival_times=spi_grb_setup1.times,
                                  measurement=spi_grb_setup1.energy_bins,
                                  n_channels=spi_grb_setup1.n_channels,
                                  start_time=spi_grb_setup1.time_start,
                                  stop_time=spi_grb_setup1.time_stop,
                                  dead_time=np.zeros_like(spi_grb_setup1.times),
                                  first_channel=0,
                                  instrument=spi_grb_setup1.det_name,
                                  mission=spi_grb_setup1.mission,
                                  verbose=False,
                                  edges=spi_grb_setup1.ebounds,
        )

        tsb = cls(f"spi-{pointing_id}-{det}",
                  e,
                  verbose=True,
                  response=response,
                  container_type=BinnedSpectrumWithDispersion)

        # set active time to total time of sw
        tsb.set_active_time_interval(f"{spi_grb_setup1.times[0]}-"
                                     f"{spi_grb_setup1.times[-1]}")

        return tsb



# Test package data

def test_packagedata():
    from pyspi.io.package_data import (get_path_of_internal_data_dir,
                                       get_path_of_external_data_dir)
    

    get_path_of_internal_data_dir()
    get_path_of_external_data_dir()



import os


def get_path_of_external_data_dir():
    """
    Get path to the external data directory (mostly to store data there)
    """
    file_path = os.environ['PYSPI']

    return file_path


def get_path_of_internal_data_dir():
    """
    Get path to the external data directory (mostly to store data there)
    """
    file_path = os.environ['PYSPI_PACKAGE_DATA']

    return file_path



import os
from shutil import copyfile
from pyspi.io.package_data import get_path_of_external_data_dir
from astropy.utils.data import download_file
import requests
import shutil
import urllib

from pyspi.io.file_utils import file_existing_and_readable

def create_file_structure(pointing_id):
    """
    Create the file structure to save the datafiles
    :param pointing_id: Id of pointing e.g. '180100610010' as string!
    :return:
    """
    # Check if file structure exists. If not, create it.
    if not os.path.exists(get_path_of_external_data_dir()):
        os.mkdir(get_path_of_external_data_dir())

    if not os.path.exists(os.path.join(get_path_of_external_data_dir(),
                                       'pointing_data')):
        os.mkdir(os.path.join(get_path_of_external_data_dir(),
                              'pointing_data'))

    if not os.path.exists(os.path.join(get_path_of_external_data_dir(),
                                       'pointing_data',
                                       pointing_id)):
        os.mkdir(os.path.join(get_path_of_external_data_dir(),
                              'pointing_data',
                              pointing_id))


def get_and_save_file(file_path, file_save_path, access="isdc"):
    """
    Function to get and save a file located at file_path to file_save_path
    :param file_path: File location (link or path to afs)
    :param file_save_path: File Save location (on local system)
    :param access: How to get the data. Possible are "isdc" and "afs"
    :return:
    """
    assert access in ["isdc", "afs"],\
        f"Access variable must be 'isdc' or 'afs' but is {access}."

    if not file_existing_and_readable(file_save_path):
        if access == "afs":
            assert os.path.exists(file_path), "Either pointing_id "\
                "is not valid, or you have no access to the afs server "\
                "or no rights to read the integral data"

            copyfile(file_path, file_save_path)

        else:

            try:
                urllib.request.urlopen(file_path)

            except:

                raise AssertionError(f'Link {file_path} does not exists!')

            data = download_file(file_path)
            shutil.move(data, file_save_path)


def get_files(pointing_id, access="isdc"):
    """
    Function to get the needed files for a certain pointing_id and save
    them in the correct folders.
    :param pointing_id: Id of pointing e.g. '180100610010' as string or int
    :param access: How to get the data. Possible are "isdc" and "afs"
    :return:
    """
    # If pointing_id is given as integer, convert it to string
    pointing_id = str(pointing_id)

    assert access in ["isdc", "afs"],\
        f"Access variable must be 'isdc' or 'afs' but is {access}."

    # Path where data should be stored
    geom_save_path = os.path.join(get_path_of_external_data_dir(),
                                  'pointing_data',
                                  pointing_id,
                                  'sc_orbit_param.fits.gz')
    data_save_path = os.path.join(get_path_of_external_data_dir(),
                                  'pointing_data',
                                  pointing_id,
                                  'spi_oper.fits.gz')
    hk_save_path = os.path.join(get_path_of_external_data_dir(),
                                'pointing_data',
                                pointing_id,
                                'spi_science_hk.fits.gz')

    if access == "afs":
        # Path to pointing_id directory
        dir_link = "/afs/ipp-garching.mpg.de/mpe/gamma/"\
            "instruments/integral/data/revolutions/"\
            "{}/{}.001/".format(pointing_id[:4], pointing_id)

    else:

        dir_link = "ftp://isdcarc.unige.ch/arc/rev_3/scw/"\
            "{}/{}.001/".format(pointing_id[:4], pointing_id)

    # Paths to the data file and the orbit file on the afs server
    geom_path = os.path.join(dir_link,
                                 'sc_orbit_param.fits.gz')
    data_path = os.path.join(dir_link,
                                 'spi_oper.fits.gz')
    hk_path = os.path.join(dir_link,
                                'spi_science_hk.fits.gz')

    create_file_structure(pointing_id)

    # Get the data files we need
    get_and_save_file(geom_path, geom_save_path, access=access)
    get_and_save_file(data_path, data_save_path, access=access)
    get_and_save_file(hk_path, hk_save_path, access=access)




def test_grb_fit():
    from pyspi.SPILike import SPILikeGRB
    from threeML import DataList
    from pyspi.utils.data_builder.time_series_builder import TimeSeriesBuilderSPI
    from pyspi.utils.response.spi_response import ResponseRMFGenerator
    from pyspi.utils.response.spi_drm import SPIDRM
    from astropy.time import Time
    from astromodels import (Powerlaw, Log_uniform_prior, Uniform_prior,
                             PointSource, Model)
    import numpy as np
    from pyspi.utils.function_utils import find_response_version
    from pyspi.utils.response.spi_response_data import ResponseDataRMF
    from threeML import BayesianAnalysis

    ein = np.geomspace(20, 800, 300)
    ebounds = np.geomspace(20, 400, 30)
    grbtime = Time("2012-07-11T02:44:53", format='isot', scale='utc')
    version = find_response_version(grbtime)
    rsp_base = ResponseDataRMF.from_version(version)
    active_time = "65-75"
    bkg_time1 = "-500--10"
    bkg_time2 = "150-1000"
    spilikes = []
    ra_val = 94.6783
    dec_val = -70.99905
    for d in [0, 12, 18]:
        drm_generator = ResponseRMFGenerator.from_time(grbtime,
                                                       d,
                                                       ebounds,
                                                       ein,
                                                       rsp_base)
        sd = SPIDRM(drm_generator, ra_val, dec_val)
        tsb = TimeSeriesBuilderSPI.from_spi_grb(f"SPIDet{d}",
                                                d,
                                                grbtime,
                                                response=sd,
                                                sgl_type="both",
                                                )
        tsb.set_active_time_interval(active_time)
        tsb.set_background_interval(bkg_time1, bkg_time2)

        sl = tsb.to_spectrumlike()
        spilikes.append(SPILikeGRB.from_spectrumlike(sl,
                                                     free_position=False))
    datalist = DataList(*spilikes)

    pl = Powerlaw()
    pl.K.prior = Log_uniform_prior(lower_bound=1e-6, upper_bound=1e4)
    pl.K.bounds = (1e-6, 1e4)
    pl.index.set_uninformative_prior(Uniform_prior)
    pl.piv.value = 200
    ps = PointSource('GRB', ra=ra_val, dec=dec_val, spectral_shape=pl)

    model = Model(ps)

    ba_spi = BayesianAnalysis(model, datalist)
    ba_spi.set_sampler("emcee", share_spectrum=True)
    ba_spi.sampler.setup(n_walkers=20, n_iterations=500, seed=1000)
    ba_spi.sample()

    expected_result = [0.02192459821838934, -1.0371965175849485]

    assert np.isclose(
        ba_spi.results.get_data_frame()["value"]
        ["GRB.spectrum.main.Powerlaw.K"],
        expected_result[0],
        rtol=0.1,
    )

    assert np.isclose(
        ba_spi.results.get_data_frame()["value"]
        ["GRB.spectrum.main.Powerlaw.index"],
        expected_result[1],
        rtol=0.1,
    )



import numpy as np
from astropy.time.core import Time, TimeDelta
from datetime import datetime
import h5py
import os

from pyspi.io.package_data import get_path_of_internal_data_dir


def get_time_object(time):
    """
    Transform the input into a time object. Input can either be a
    time object or a string with the format "YYMMDD HHMMSS"

    :param time: time object or a string with the format "YYMMDD HHMMSS"

    :returns: time object
    """
    if not isinstance(time, Time):
        time = datetime.strptime(time,
                                 '%y%m%d %H%M%S')
        time = Time(time)
    return time


def leapseconds(time_object):
    """
    Hard coded leap seconds from start of INTEGRAL to time of time_object

    :param time_object: Time object to which the number of
    leapseconds should be detemined

    :returns: TimeDelta object of the needed leap seconds
    """

    time_object = get_time_object(time_object)

    if time_object < Time(datetime.strptime('060101 000000',
                                            '%y%m%d %H%M%S')):
        lsec = 0
    elif time_object < Time(datetime.strptime('090101 000000',
                                              '%y%m%d %H%M%S')):
        lsec = 1
    elif time_object < Time(datetime.strptime('120701 000000',
                                              '%y%m%d %H%M%S')):
        lsec = 2
    elif time_object < Time(datetime.strptime('150701 000000',
                                              '%y%m%d %H%M%S')):
        lsec = 3
    elif time_object < Time(datetime.strptime('170101 000000',
                                              '%y%m%d %H%M%S')):
        lsec = 4
    else:
        lsec = 5
    return TimeDelta(lsec, format='sec')


def find_response_version(time):
    """
    Find the correct response version number for a given time

    :param time: time of interest

    :returns: response version number
    """

    time = get_time_object(time)

    if time < Time(datetime.strptime('031206 060000', '%y%m%d %H%M%S')):
        version = 0
    elif time < Time(datetime.strptime('040717 082006', '%y%m%d %H%M%S')):
        version = 1
    elif time < Time(datetime.strptime('090219 095957', '%y%m%d %H%M%S')):
        version = 2
    elif time < Time(datetime.strptime('100527 124500', '%y%m%d %H%M%S')):
        version = 3
    else:
        version = 4
    return version

def find_needed_ids(time):
    """
    Get the pointing id of the needed data to cover the GRB time

    :returns: Needed pointing id
    """

    time = get_time_object(time)

    # Path to file, which contains id information and start and stop
    # time
    id_file_path = os.path.join(get_path_of_internal_data_dir(),
                                'id_data_time.hdf5')

    # Get GRB time in ISDC_MJD
    time_of_GRB_ISDC_MJD = (time).tt.mjd-51544
    # Get which id contain the needed time. When the wanted time is
    # too close to the boundarie also add the pervious or following
    # observation id
    id_file = h5py.File(id_file_path, 'r')
    start_id = id_file['Start'][()]
    stop_id = id_file['Stop'][()]
    ids = id_file['ID'][()]

    mask_larger = start_id < time_of_GRB_ISDC_MJD
    mask_smaller = stop_id > time_of_GRB_ISDC_MJD

    try:
        id_number = list(mask_smaller*mask_larger).index(True)
    except:
        raise Exception('No pointing id contains this time...')

    return ids[id_number].decode("utf-8")


def ISDC_MJD(time_object):
    """
    Get INTEGRAL MJD time from a given time object

    :param time_object: Astropy time object of grb time

    :returns: Time in Integral MJD time
    """
    time_object = get_time_object(time_object)
    return time_object.tt.mjd-51544


def ISDC_MJD_to_cxcsec(ISDC_MJD_time):
    """
    Convert ISDC_MJD to UTC

    :param ISDC_MJD_time: time in ISDC_MJD time format

    :returns: time in cxcsec format (seconds since 1998-01-01 00:00:00)
    """

    return Time(ISDC_MJD_time+51544, format='mjd', scale='utc').cxcsec



import numpy as np
import h5py
from dataclasses import dataclass
import os
import astropy.io.fits as fits

from pyspi.io.package_data import get_path_of_internal_data_dir


def load_rmf_non_ph_1():
    """
    Load the RMF for the non-photopeak events that first interact in the det

    :returns: ebounds of RMF and rmf matrix for the non-photopeak events that
        first interact in the det
    """

    with fits.open(os.path.join(
            get_path_of_internal_data_dir(),
            'spi_rmf2_rsp_0002.fits')
                   ) as rmf_file:
        rmf_comp = rmf_file['SPI.-RMF2-RSP'].data['MATRIX']
        emax = rmf_file['SPI.-RMF2-RSP'].data['ENERG_HI']
        emin = rmf_file['SPI.-RMF2-RSP'].data['ENERG_LO']

    ebounds = np.append(emin, emax[-1])

    # The RMFs are stored in a weird way, we have to expand this to a
    # real square matrix
    rmf = np.zeros((len(rmf_comp), len(rmf_comp)))
    for i, row in enumerate(rmf_comp):
        length = len(row.flatten())
        rmf[i, :length] = row.flatten()

    return ebounds, rmf


def load_rmf_non_ph_2():
    """
    Load the RMF for the non-photopeak events that first interact
    in the dead material

    :returns: ebounds of RMF and rmf matrix for the non-photopeak events that
        first interact in the dead material
    """

    with fits.open(os.path.join(
            get_path_of_internal_data_dir(),
            'spi_rmf3_rsp_0002.fits')
                   ) as rmf_file:
        rmf_comp = rmf_file['SPI.-RMF3-RSP'].data['MATRIX']
        emax = rmf_file['SPI.-RMF3-RSP'].data['ENERG_HI']
        emin = rmf_file['SPI.-RMF3-RSP'].data['ENERG_LO']

    ebounds = np.append(emin, emax[-1])

    # The RMFs are stored in a weird way, we have to expand this to a
    # real square matrix
    rmf = np.zeros((len(rmf_comp), len(rmf_comp)))
    for i, row in enumerate(rmf_comp):
        length = len(row.flatten())
        rmf[i, :length] = row.flatten()
    rmf[0] = np.zeros(len(rmf))
    return ebounds, rmf


@dataclass
class ResponseData:
    """
    Base Dataclass to hold the IRF data
    """

    energies_database: np.array
    irf_xmin: float
    irf_ymin: float
    irf_xbin: float
    irf_ybin: float
    irf_nx: int
    irf_ny: int
    n_dets: int
    ebounds_rmf_2_base: np.array
    rmf_2_base: np.array
    ebounds_rmf_3_base: np.array
    rmf_3_base: np.array

    def get_data(self, version):
        """
        Read in the data we need from the irf hdf5 file

        :param version: Version of irf file

        :returns: all the infomation we need as a list
        """
        assert version in [0, 1, 2, 3, 4],\
            f"Version must be in [0, 1, 2, 3, 4] but is {version}"

        irf_file = os.path.join(get_path_of_internal_data_dir(),
                                f"spi_three_irfs_database_{version}.hdf5")

        if version == 0:
            print('Using the irfs that are valid between Start'
                  ' and 03/07/06 06:00:00 (YY/MM/DD HH:MM:SS)')

        elif version == 1:
            print('Using the irfs that are valid between 03/07/06 06:00:00'
                  ' and 04/07/17 08:20:06 (YY/MM/DD HH:MM:SS)')

        elif version == 2:
            print('Using the irfs that are valid between 04/07/17 08:20:06'
                  ' and 09/02/19 09:59:57 (YY/MM/DD HH:MM:SS)')

        elif version == 3:
            print('Using the irfs that are valid between 09/02/19 09:59:57'
                  ' and 10/05/27 12:45:00 (YY/MM/DD HH:MM:SS)')

        else:
            print('Using the irfs that are valid between 10/05/27 12:45:00'
                  ' and present (YY/MM/DD HH:MM:SS)')

        irf_database = h5py.File(irf_file, 'r')
        energies_database = irf_database['energies'][()]
        irf_data = irf_database['irfs']
        irfs = irf_data[()]
        irf_xmin = irf_data.attrs['irf_xmin']
        irf_ymin = irf_data.attrs['irf_ymin']
        irf_xbin = irf_data.attrs['irf_xbin']
        irf_ybin = irf_data.attrs['irf_ybin']
        irf_nx = irf_data.attrs['nx']
        irf_ny = irf_data.attrs['ny']
        irf_database.close()

        ebounds_rmf_2_base, rmf_2_base = load_rmf_non_ph_1()
        ebounds_rmf_3_base, rmf_3_base = load_rmf_non_ph_2()

        return irfs, energies_database, irf_xmin, irf_ymin, irf_xbin, \
            irf_ybin, irf_nx, irf_ny, irf_data, ebounds_rmf_2_base, \
            rmf_2_base, ebounds_rmf_3_base, rmf_3_base


@dataclass
class ResponseDataPhotopeak(ResponseData):
    """
    Dataclass to hold the IRF data if we only need the photopeak irf
    """
    irfs_photopeak: np.array

    @classmethod
    def from_version(cls, version):
        """
        Construct the dataclass object

        :param version: Which IRF version?

        :returns: ResponseDataPhotopeak object
        """
        data = super().get_data(ResponseData, version)

        irfs_photopeak = data[0][:, :, :, :, 0]

        return cls(*data[1:],
                   irfs_photopeak)


@dataclass
class ResponseDataRMF(ResponseData):
    """
    Dataclass to hold the IRF data if we only need all three irfs
    """
    irfs_photopeak: np.array
    irfs_nonphoto_1: np.array
    irfs_nonphoto_2: np.array

    @classmethod
    def from_version(cls, version):
        """
        Construct the dataclass object

        :param version: Which IRF version?

        :returns: ResponseDataPhotopeak object
        """
        data = super().get_data(ResponseData, version)

        irfs_photopeak = data[0][:, :, :, :, 0]
        irfs_nonphoto_1 = data[0][:, :, :, :, 1]
        irfs_nonphoto_2 = data[0][:, :, :, :, 2]

        return cls(*data[1:],
                   irfs_photopeak,
                   irfs_nonphoto_1,
                   irfs_nonphoto_2)



import numpy as np


def test_active_dets_and_response_version():
    from pyspi.utils.livedets import get_live_dets
    from pyspi.utils.function_utils import find_response_version

    time_string = "051212 205010"
    ld = get_live_dets(time_string, event_types="single")
    version = find_response_version(time_string)

    assert np.sum(ld) == 152, f"Got wrong active det numbers for {time_string}"
    assert version == 2, f"Got wrong rsp version number for {time_string}"

    time_string = "031212 205010"
    ld = get_live_dets(time_string, event_types="single")
    version = find_response_version(time_string)

    assert np.sum(ld) == 169, f"Got wrong active det numbers for {time_string}"
    assert version == 1, f"Got wrong rsp version number for {time_string}"

    time_string = "151212 205010"
    ld = get_live_dets(time_string, event_types="single")
    version = find_response_version(time_string)

    assert np.sum(ld) == 146, f"Got wrong active det numbers for {time_string}"
    assert version == 4, f"Got wrong rsp version number for {time_string}"


def test_plotting():
    from pyspi.io.plotting.spi_display import SPI
    time_string = "151212 205010"
    s = SPI(time=time_string)
    fig = s.plot_spi_working_dets()



import numpy as np
import astropy.io.fits as fits
from numba import njit
import os

from pyspi.utils.geometry import cart2polar, polar2cart
from pyspi.io.package_data import get_path_of_internal_data_dir


@njit
def construct_scy(scx_ra, scx_dec, scz_ra, scz_dec):
    """
    Construct the vector of the y-axis of the Integral coord system
    in the ICRS frame

    :param scx_ra: ra coordinate of satellite x-axis in ICRS
    :param scx_dec: dec coordinate of satellite x-axis in ICRS
    :param scz_ra: ra coordinate of satellite z-axis in ICRS
    :param scz_dec: dec coordinate of satellite z-axis in ICRS

    :returns: vector of the y-axis of the Integral coord system
    in the ICRS frame
    """
    x = polar2cart(scx_ra, scx_dec)
    z = polar2cart(scz_ra, scz_dec)

    return cart2polar(np.cross(x, -z))


@njit
def construct_sc_matrix(scx_ra, scx_dec, scy_ra, scy_dec, scz_ra, scz_dec):
    """
    Construct the sc_matrix, with which we can transform ICRS <-> Sat. Frame

    :param scx_ra: ra coordinate of satellite x-axis in ICRS
    :param scx_dec: dec coordinate of satellite x-axis in ICRS
    :param scy_ra: ra coordinate of satellite y-axis in ICRS
    :param scy_dec: dec coordinate of satellite y-axis in ICRS
    :param scz_ra: ra coordinate of satellite z-axis in ICRS
    :param scz_dec: dec coordinate of satellite z-axis in ICRS

    :returns: sc_matrix (3x3)
    """

    sc_matrix = np.zeros((3, 3))

    sc_matrix[0, :] = polar2cart(scx_ra, scx_dec)
    sc_matrix[1, :] = polar2cart(scy_ra, scy_dec)
    sc_matrix[2, :] = polar2cart(scz_ra, scz_dec)

    return sc_matrix


class SPIPointing:

    def __init__(self, sc_pointing_file):
        """
        This class handles the **current** SPI pointings
        based of the input SPI pointing file

        :param sc_pointing_file: An INTEGRAL/SPI spacecraft pointing file

        :returns:
        """

        # construct the misalignment matrix
        self._open_misalignment_matrix()

        # open the space craft pointing file and extract it
        with fits.open(sc_pointing_file) as f:

            self._pointing_data = f['INTL-ORBI-SCP'].data

        # construct the space craft pointings

        self._construct_sc_matrices()


    def _open_misalignment_matrix(self):
        """
        Open and build the INTEGRAL to SPI misalignment matrix
        that corrects SPI's pointing to the full INTEGRAL
        pointing

        :returns:
        """

        # get the path to the data file
        matrix_file = os.path.join(get_path_of_internal_data_dir(),
                                   'inst_misalign_20050328.fits')

        # open the file
        with fits.open(matrix_file) as f:

            # SPI should always be at idx == 2, but lets make sure!
            spi_idx = f['GNRL-IROT-MOD'].data['INSTRUMENT'] == 'SPI'

            # extract the raw matrix
            matrix_raw = f['GNRL-IROT-MOD'].data['MATRIX'][spi_idx]

            # now reshape it
            self._misalignment_matrix = matrix_raw.reshape((3,3))

    def _construct_sc_matrices(self):
        """
        Extract and construct the SPI pointings taking into account the
        misalignment between INTEGRAL and SPI

        :returns:
        """

        self._n_pointings = len(self._pointing_data)

        self._sc_matrix = np.zeros((self._n_pointings,3,3))
        self._sc_points = []

        # Look through all times and construct the pointing matrix
        for i in range(self._n_pointings):

            # first wwe construct the matric for INTEGRAL
            scx_ra = self._pointing_data['RA_SCX'][i]
            scx_dec = self._pointing_data['DEC_SCX'][i]

            scz_ra = self._pointing_data['RA_SCZ'][i]
            scz_dec = self._pointing_data['DEC_SCZ'][i]

            scy_ra, scy_dec = construct_scy(scx_ra, scx_dec, scz_ra, scz_dec)

            intergal_matrix = construct_sc_matrix(scx_ra, scx_dec, scy_ra,
                                                   scy_dec, scz_ra, scz_dec)

            # now apply the misalignment matrix
            spi_matrix = np.dot(self._misalignment_matrix, intergal_matrix)

            # now convert the ra and dec to the proper frame
            scx_ra, scx_dec = cart2polar(spi_matrix[0])
            scy_ra, scy_dec = cart2polar(spi_matrix[1])
            scz_ra, scz_dec = cart2polar(spi_matrix[2])

            self._sc_matrix[i, ...] = spi_matrix
            self._sc_points.append(dict(scx_ra=scx_ra, scx_dec=scx_dec,
                                        scy_ra=scy_ra, scy_dec=scy_dec,
                                        scz_ra=scz_ra, scz_dec=scz_dec))

    @property
    def sc_matrix(self):
        """
        get the sc_matrics of all the times in this pointing
        :returns: array of sc_matrices
        """
        return self._sc_matrix

    @property
    def sc_points(self):
        """
        ra, dec coordinates of the SPI x,y and z axis in the ICRS frame
        for all the times in this pointing

        :returns: ra, dec coordinates of the SPI x,y and z axis in the \
        ICRS frame for all the pointings
        """
        return self._sc_points
        



#from pyspi.spi_response import SPIResponse
#from pyspi.SPILike import SPILike


#from IPython.display import display#, Video



#display(Video('https://br-i.akamaihd.net/i/MUJIuUOVBwQIbtC2uKJDM6OhuLnC_2rc5K1S/_AJS/9-rc5ygy/4ff8be83-80eb-45e6-bf4c-2b43e2344759_,0,A,B,E,C,X,.mp4.csmil/master.m3u8?__b__=200'))

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np


def test_response():
    from astropy.time import Time
    from pyspi.utils.function_utils import find_response_version
    from pyspi.utils.response.spi_response_data import ResponseDataRMF
    from pyspi.utils.response.spi_response import ResponseRMFGenerator
    from pyspi.utils.response.spi_drm import SPIDRM

    grbtime = Time("2012-07-11T02:44:53", format='isot', scale='utc')
    ein = np.geomspace(20,8000,300)
    ebounds = np.geomspace(20,4000,30)

    version = find_response_version(grbtime)
    rsp_base = ResponseDataRMF.from_version(version)

    # Check the rsp_base values
    assert np.isclose(rsp_base.irfs_nonphoto_1.sum(), 1720731.0176313636), \
        "The IRF values in the rsp base object are not the ones we expect."

    assert np.isclose(rsp_base.ebounds_rmf_3_base.sum(), 74573.8), \
        "The Ebounds in the rsp base object are not the ones we expect."

    det = 0
    ra_val = 94.6783
    dec_val = -70.99905
    drm_generator = ResponseRMFGenerator.from_time(grbtime,
                                                   det,
                                                   ebounds,
                                                   ein,
                                                   rsp_base)
    sd = SPIDRM(drm_generator, ra_val, dec_val)

    assert sd.matrix.sum() == 1846.700886770058, \
        "We expect a different response matrix..."



import numpy as np
from numba import njit

@njit
def polar2cart(ra, dec):
    """
    Convert ra, dec to cartesian

    :param ra: ra coord
    :param dec: dec coord

    :returns: cartesian coord vector
    """
    x = np.cos(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
    y = np.sin(np.deg2rad(ra)) * np.cos(np.deg2rad(dec))
    z = np.sin(np.deg2rad(dec))

    return np.array([x,y,z])


@njit
def cart2polar(vector):
    """
    Convert cartesian to ra, dec

    :param vector: cartesian coord vector

    :returns: ra and dec
    """
    ra = np.arctan2(vector[1], vector[0])
    dec = np.arcsin(vector[2])

    return np.rad2deg(ra), np.rad2deg(dec)



# Test download from isdarc for given pointin_id

def test_download():
    from pyspi.io.get_files import get_files
    from pyspi.io.package_data import get_path_of_external_data_dir
    from pyspi.io.file_utils import file_existing_and_readable
    import os

    pointing_id = 169600130010
    get_files(pointing_id)

    geom_save_path = os.path.join(get_path_of_external_data_dir(),
                                  'pointing_data',
                                  str(pointing_id),
                                  'sc_orbit_param.fits.gz')
    data_save_path = os.path.join(get_path_of_external_data_dir(),
                                  'pointing_data',
                                  str(pointing_id),
                                  'spi_oper.fits.gz')

    assert file_existing_and_readable(geom_save_path), 'Download test failed'
    assert file_existing_and_readable(data_save_path), 'Download test failed'


    



import astropy.coordinates as coord
import astropy.units as u
import numpy as np
from numba import njit
from astropy.coordinates import (BaseCoordinateFrame,
                                 Attribute,
                                 RepresentationMapping,
                                 frame_transform_graph,
                                 spherical_to_cartesian)

from pyspi.utils.response.spi_pointing import construct_sc_matrix
from pyspi.utils.geometry import cart2polar, polar2cart

class SPIFrame(BaseCoordinateFrame):
    """
    
    INTEGRAL SPI Frame
    Parameters
    ----------
    representation : `BaseRepresentation` or None
        A representation object or None to have no data (or use the other keywords)
  
    """
    default_representation = coord.SphericalRepresentation

    frame_specific_representation_info = {
        'spherical': [
            RepresentationMapping(
                reprname='lon', framename='lon', defaultunit=u.degree),
            RepresentationMapping(
                reprname='lat', framename='lat', defaultunit=u.degree),
            RepresentationMapping(
                reprname='distance', framename='DIST', defaultunit=None)
        ],
        'unitspherical': [
            RepresentationMapping(
                reprname='lon', framename='lon', defaultunit=u.degree),
            RepresentationMapping(
                reprname='lat', framename='lat', defaultunit=u.degree)
        ],
        'cartesian': [
            RepresentationMapping(
                reprname='x', framename='SCX'), RepresentationMapping(
                reprname='y', framename='SCY'), RepresentationMapping(
                reprname='z', framename='SCZ')
        ]
    }

    # Specify frame attributes required to fully specify the frame
    scx_ra = Attribute(default=None)
    scx_dec = Attribute(default=None)
    
    scy_ra = Attribute(default=None)
    scy_dec = Attribute(default=None)
    
    scz_ra = Attribute(default=None)
    scz_dec = Attribute(default=None)
   




@frame_transform_graph.transform(coord.FunctionTransform, SPIFrame, coord.ICRS)
def spi_to_j2000(spi_coord, j2000_frame):
    """ 
    Transform spi fram to ICRS frame
    """

    sc_matrix = construct_sc_matrix(spi_coord.scx_ra,
                                    spi_coord.scx_dec,
                                    spi_coord.scy_ra,
                                    spi_coord.scy_dec,
                                    spi_coord.scz_ra,
                                    spi_coord.scz_dec)

    # X,Y,Z = gbm_coord.cartesian

    pos = spi_coord.cartesian.xyz.value

    X0 = np.dot(sc_matrix[:, 0], pos)
    X1 = np.dot(sc_matrix[:, 1], pos)
    X2 = np.clip(np.dot(sc_matrix[:, 2], pos), -1., 1.)

    dec = np.pi / 2. - np.arccos(X2)# np.arcsin(X2)

    idx = np.logical_and(np.abs(X0) < 1E-6, np.abs(X1) < 1E-6)

    ra = np.zeros_like(dec)

    ra[~idx] = np.arctan2(X1[~idx], X0[~idx]) % (2 * np.pi)

    return coord.ICRS(ra=ra * u.radian, dec=dec * u.radian)


@frame_transform_graph.transform(coord.FunctionTransform, coord.ICRS, SPIFrame)
def j2000_to_spi(j2000_frame, spi_coord):
    """ 
    Transform icrs frame to SPI frame
    """

    sc_matrix = construct_sc_matrix(spi_coord.scx_ra,
                                    spi_coord.scx_dec,
                                    spi_coord.scy_ra,
                                    spi_coord.scy_dec,
                                    spi_coord.scz_ra,
                                    spi_coord.scz_dec)
  

    pos = j2000_frame.cartesian.xyz.value

    X0 = np.dot(sc_matrix[0, :], pos)
    X1 = np.dot(sc_matrix[1, :], pos)
    X2 = np.dot(sc_matrix[2, :], pos)

    lat = np.pi / 2. - np.arccos(X2)  # convert to proper frame

    idx = np.logical_and(np.abs(X0) < 1E-6, np.abs(X1) < 1E-6)

    lon = np.zeros_like(lat)

    lon[~idx] = np.arctan2(X1[~idx], X0[~idx]) % (2 * np.pi)

    return SPIFrame(
        lon=lon * u.radian,
        lat=lat * u.radian,
        scx_ra=spi_coord.scx_ra,
        scx_dec=spi_coord.scx_dec,
        scy_ra=spi_coord.scy_ra,
        scy_dec=spi_coord.scy_dec,
        scz_ra=spi_coord.scz_ra,
        scz_dec=spi_coord.scz_dec

    )

# Functions to do the coordinate transformation fast! But has none
# of the astropy benefits.
@njit
def transform_icrs_to_spi(ra_icrs, dec_icrs, sc_matrix):
    """
    Calculates lon, lat in spi frame for given ra, dec in ICRS frame and given
    sc_matrix (sc_matrix pointing dependent)

    :param ra_icrs: Ra in ICRS in degree
    :param dec_icrs: Dec in ICRS in degree
    :param sc_matrix: sc Matrix that gives orientation of SPI in ICRS frame

    :returns: lon, lat in spi frame
    """
    # Source in icrs
    vec_ircs = polar2cart(ra_icrs, dec_icrs)
    vec_spi = np.dot(sc_matrix, vec_ircs)
    lon, lat = cart2polar(vec_spi)
    if lon < 0:
        lon += 360
    return lon, lat

@njit
def transform_spi_to_icrs(az_spi, zen_spi, sc_matrix):
    """
    Calculates lon, lat in spi frame for given ra, dec in ICRS frame and given
    sc_matrix (sc_matrix pointing dependent)

    :param az_spi: azimuth in SPI coord system in degree
    :param zen_spi: zenit in SPI coord system in degree
    :param sc_matrix: sc Matrix that gives orientation of SPI in ICRS frame

    :returns: ra, dex in ICRS in deg
    """
    # Source in icrs
    vec_spi = polar2cart(az_spi, zen_spi)
    vec_icrs = np.dot(np.linalg.inv(sc_matrix), vec_spi)
    ra, dec = cart2polar(vec_icrs)
    if ra < 0:
        ra += 360
    return ra, dec



import pytest


def test_time_series_without_response():
    from pyspi.utils.data_builder.time_series_builder import \
        TimeSeriesBuilderSPI
    import numpy as np
    from astropy.time import Time
    ebounds = np.geomspace(20, 400, 30)
    grbtime = Time("2012-07-11T02:44:53", format='isot', scale='utc')
    tsb = TimeSeriesBuilderSPI.from_spi_grb("SPIDet0",
                                            0,
                                            grbtime,
                                            ebounds=ebounds,
                                            sgl_type="both",
                                            )
    assert np.isclose(tsb.time_series.
                      count_per_channel_over_interval(-500, 500)[10],
                      2120)


def test_time_series_with_response():
    from pyspi.utils.data_builder.time_series_builder import \
        TimeSeriesBuilderSPI
    import numpy as np
    from astropy.time import Time
    from pyspi.utils.function_utils import find_response_version
    from pyspi.utils.response.spi_response_data import ResponseDataRMF
    from pyspi.utils.response.spi_response import ResponseRMFGenerator
    from pyspi.utils.response.spi_drm import SPIDRM
    ebounds = np.geomspace(20, 400, 30)
    ein = np.geomspace(20, 800, 300)
    grbtime = Time("2012-07-11T02:44:53", format='isot', scale='utc')
    version = find_response_version(grbtime)
    rsp_base = ResponseDataRMF.from_version(version)
    ra_val = 94.6783
    dec_val = -70.99905
    drm_generator = ResponseRMFGenerator.from_time(grbtime,
                                                   0,
                                                   ebounds,
                                                   ein,
                                                   rsp_base)
    sd = SPIDRM(drm_generator, ra_val, dec_val)
    tsb = TimeSeriesBuilderSPI.from_spi_grb("SPIDet0",
                                            0,
                                            grbtime,
                                            response=sd,
                                            sgl_type="both",
                                            )

    assert np.isclose(tsb.time_series.
                      count_per_channel_over_interval(-500, 500)[10],
                      2120)

    tsb = TimeSeriesBuilderSPI.from_spi_grb("SPIDet0",
                                            0,
                                            grbtime,
                                            ebounds=ebounds,
                                            response=sd,
                                            sgl_type="both",
                                            )
    assert np.isclose(tsb.time_series.
                      count_per_channel_over_interval(-500, 500)[10],
                      2120)

    with pytest.raises(AssertionError):
        tsb = TimeSeriesBuilderSPI.from_spi_grb("SPIDet0",
                                                0,
                                                grbtime,
                                                ebounds=ein,
                                                response=sd,
                                                sgl_type="both",
                                                )



import numpy as np
from threeML.utils.OGIP.response import InstrumentResponse


class SPIDRM(InstrumentResponse):
    def __init__(self, drm_generator, ra, dec):
        """
        Init a SPIDRM object which is based on the InstrumenResponse
        class from 3ML. Contains everything that is necessary for
        3ML to recognize it as a response.

        :param drm_generator: DRM generator for the SPI Response
        :param ra: ra of source (in ICRS)
        :param dec: dec of source (in ICRS)
        """
        self._drm_generator = drm_generator

        self._drm_generator.set_location(ra, dec)
        self._min_dist = np.deg2rad(.5)

        self._ra = ra
        self._dec = dec

        super(SPIDRM, self).__init__(
            self._drm_generator.matrix,
            self._drm_generator.ebounds,
            self._drm_generator.monte_carlo_energies,
        )

    def set_location(self, ra, dec, cache=False):
        """
        Set the source location

        :param ra: ra of source (in ICRS)
        :param dec: dec of source (in ICRS)

        :returns:
        """
        self._drm_generator.set_location(ra, dec)

        self._ra = ra
        self._dec = dec

        self._matrix = self._drm_generator.matrix
        self._matrix_transpose = self._matrix.T

    def set_location_direct_sat_coord(self, azimuth, zenith, cache=False):
        """
        Set the source location

        :param azimuth: az poisition in the sat. frame
        :param zenith: zenith poisition in the sat. frame

        :returns:
        """
        self._ra, self._dec =\
            self._drm_generator.set_location_direct_sat_coord(azimuth, zenith)

        self._matrix = self._drm_generator.matrix
        self._matrix_transpose = self._matrix.T

    def clone(self) -> "SPIDRM":
        """
        Get clone of this response object

        :returns: new cloned response
        """

        # We have to be carefull to not clone the irf_read_object.
        # Because it is huge! ~ 1GB
        new_drm_generator = self._drm_generator.clone()

        #    ResponseRMFNew(monte_carlo_energies=self._drm_generator.monte_carlo_energies,
        #                   ebounds=self._drm_generator.ebounds,
        #                   response_irf_read_object=self._drm_generator._irf_ob,
        #                   sc_matrix=self._drm_generator._sc_matrix,
        #                   det=self._drm_generator._det,
        #                   fixed_rsp_matrix=self._drm_generator._rsp_matrix)
        return SPIDRM(drm_generator=new_drm_generator,
                      ra=self._ra,
                      dec=self._dec,
                      )



import collections
from typing import Optional
import numpy as np

from astromodels import Parameter, Model
from astromodels.functions.priors import Cosine_Prior, Uniform_prior

from threeML import PluginPrototype
from threeML.io.file_utils import sanitize_filename
from threeML.plugins.DispersionSpectrumLike import DispersionSpectrumLike
from threeML.plugins.SpectrumLike import SpectrumLike
from threeML.io.logging import setup_logger

from pyspi.utils.response.spi_drm import SPIDRM

log = setup_logger(__name__)


class SPILike(DispersionSpectrumLike):
    """
    Plugin for the data of SPI, based on PySPI
    """
    def __init__(
            self,
            name: str,
            observation,
            background,
            bkg_base_array,
            free_position: bool,
            verbose: bool = True,
            **kwargs
    ):
        """
        Init the plugin for a constant source analysis with PySPI

        :param name: Name of plugin
        :param observation: observed spectrum
        :param background: background spectrum
        :param bkg_base_array: Base array for background model
        :param free_position: Free the position in the fit?
        :param verbose: Verbose?

        :returns: Object
        """
        self._free_position: bool = free_position

        if not isinstance(
                observation.response, SPIDRM
        ):

            log.error("The response associated with the observation"
                      " is not a SPIDRM")

            raise AssertionError()

        super(SPILike, self).__init__(name,
                                      observation,
                                      background,
                                      verbose,
                                      **kwargs)

        self._bkg_base_array = bkg_base_array
        self._bkg_array = np.ones(len(self._bkg_base_array))

    def set_model(self, likelihood_model: Model) -> None:
        """
        Set the model to be used in the joint minimization.

        :param likelihood_model: likelihood model instance

        :returns:
        """

        super(SPILike, self).set_model(likelihood_model)

        if self._free_position:
            log.info(f"Freeing the position of {self.name} and setting priors")
            
            for key in self._like_model.point_sources.keys():

                self._like_model.point_sources[key].position.ra.free = True
                self._like_model.point_sources[key].position.dec.free = True

                self._like_model.point_sources[key].position.ra.prior = \
                    Uniform_prior(lower_bound=0.0, upper_bound=360)
                self._like_model.point_sources[key].position.dec.prior = \
                    Cosine_Prior(lower_bound=-90.0, upper_bound=90)

                ra = self._like_model.point_sources[key].position.ra.value
                dec = self._like_model.point_sources[key].position.dec.value
        else:

            for key in self._like_model.point_sources.keys():

                ra = self._like_model.point_sources[key].position.ra.value
                dec = self._like_model.point_sources[key].position.dec.value

        self._response.set_location(ra, dec)

    def _evaluate_model(self, precalc_fluxes=None):
        """
        Evaluate the model

        :param precalc_fluxes: Precaclulated flux of spectrum

        :returns: model counts
        """

        source = super(SPILike, self)._evaluate_model(precalc_fluxes=
                                                      precalc_fluxes)
        self._update_bkg_array()
        bkg = self._bkg_array*self._bkg_base_array
        return source+bkg

    def get_model(self, precalc_fluxes: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Get the model

        :param precalc_fluxes: Precaclulated flux of spectrum

        :returns: model counts
        """
        if self._free_position:

            # assumes that the is only one point source which is how
            # it should be!
            ra, dec = self._like_model.get_point_source_position(0)

            self._response.set_location(ra, dec)

        return super(SPILike, self).get_model(precalc_fluxes=precalc_fluxes)

    def _add_bkg_nuisance_parameter(self, bkg_parameters) -> None:
        """
        Add the bkg parameter. Are saved as array.

        :param bkg_parameters:

        :returns:
        """
        
        self._bkg_parameters = bkg_parameters
        for parameter in bkg_parameters:
            self.nuisance_parameters[parameter.name] = parameter
        self._bkg_array = np.ones(len(bkg_parameters))

    def _update_bkg_array(self) -> None:
        """
        Update the array with the background parameter

        :returns:
        """
        
        for key in self._like_model.parameters.keys():
            if "bkg" in key:
                idx = int(key.split("_")[-1])
                self._bkg_array[idx] = self._like_model.parameters[key].value

    def set_free_position(self, flag):
        """
        Set the free position flag

        :param flag: True or False

        :returns:
        """
        self._free_position = flag

    @classmethod
    def from_spectrumlike(
        cls,
        spectrum_like,
        bkg_base_array,
        free_position=False
    ):
        """
        Generate SPILikeGRB from an existing SpectrumLike child

        :param spectrum_like: SpectrumLike child
        :param rsp_object: Response object
        :free_position: Free the position? boolean

        :returns: Initialized Object
        """
        return cls(
            spectrum_like.name,
            spectrum_like._observed_spectrum,
            spectrum_like._background_spectrum,
            bkg_base_array,
            free_position,
            spectrum_like._verbose,
        )


class SPILikeGRB(DispersionSpectrumLike):
    """
    Plugin for the data of SPI, based on PySPI
    """
    def __init__(
            self,
            name,
            observation,
            background=None,
            free_position=False,
            verbose=True,
            **kwargs
    ):
        """
        Init the plugin for a GRB analysis with PySPI

        :param name: Name of plugin
        :param observation: observed spectrum
        :param background: background spectrum
        :param free_position: Free the position in the fit?
        :param verbose: Verbose?
        """

        self._free_position = free_position

        assert isinstance(
                observation.response, SPIDRM
            ), "The response associated with the observation is not a SPIDRM"

        super(SPILikeGRB, self).__init__(name,
                                         observation,
                                         background,
                                         verbose,
                                         **kwargs)

    def set_model(self, likelihood_model):
        """
        Set the model to be used in the joint minimization.

        :param likelihood_model: likelihood model instance

        :returns:
        """

        super(SPILikeGRB, self).set_model(likelihood_model)

        if self._free_position:
            print("Freeing the position of %s and setting priors" % self.name)
            for key in self._like_model.point_sources.keys():
                self._like_model.point_sources[key].position.ra.free = True
                self._like_model.point_sources[key].position.dec.free = True

                self._like_model.point_sources[key].position.ra.prior = \
                    Uniform_prior(lower_bound=0.0, upper_bound=360)
                self._like_model.point_sources[key].position.dec.prior = \
                    Cosine_Prior(lower_bound=-90.0, upper_bound=90)

                ra = self._like_model.point_sources[key].position.ra.value
                dec = self._like_model.point_sources[key].position.dec.value
        else:

            for key in self._like_model.point_sources.keys():

                ra = self._like_model.point_sources[key].position.ra.value
                dec = self._like_model.point_sources[key].position.dec.value

        self._response.set_location(ra, dec)

    def get_model(self, precalc_fluxes=None):
        """
        Get the model

        :param precalc_fluxes: Precaclulated flux of spectrum

        :returns: model counts
        """
        if self._free_position:

            # assumes that the is only one point source which is how
            # it should be!
            ra, dec = self._like_model.get_point_source_position(0)

            self._response.set_location(ra, dec)

        return super(SPILikeGRB, self).get_model(precalc_fluxes=precalc_fluxes)

    def set_free_position(self, flag):
        """
        Set the free position flag

        :param flag: True or False

        :returns:
        """
        self._free_position = flag

    @classmethod
    def from_spectrumlike(
            cls, spectrum_like, free_position=False
    ):
        """
        Generate SPILikeGRB from an existing SpectrumLike child

        :param spectrum_like: SpectrumLike child
        :param rsp_object: Response object
        :free_position: Free the position? boolean

        :returns: Initialized Object
        """
        return cls(
            spectrum_like.name,
            spectrum_like._observed_spectrum,
            spectrum_like._background_spectrum,
            free_position,
            spectrum_like._verbose,
        )



import os


def file_existing_and_readable(filename):
    """
    Check if a file exists
    :param filename: Filename to check
    :return: True or False
    """
    sanitized_filename = sanitize_filename(filename)

    if os.path.exists(sanitized_filename):

        # Try to open it

        try:

            with open(sanitized_filename):

                pass

            return True
        except FileNotFoundError:
            pass
    return False


def path_exists_and_is_directory(path):
    """
    Check if a path exists and is a directory
    :param path: Path to check
    :return: True or False
    """
    sanitized_path = sanitize_filename(path, abspath=True)

    if os.path.exists(sanitized_path):

        if os.path.isdir(path):

            return True

    return False


def sanitize_filename(filename, abspath=False):
    """
    Sanitize filename
    :param filename: name of file
    :param abspath: Get the absolute path?
    :return: sanitized filename
    """

    sanitized = os.path.expandvars(os.path.expanduser(filename))

    if abspath:

        return os.path.abspath(sanitized)

    return sanitized



import os
from datetime import datetime
import numpy as np
import scipy.interpolate as interpolate
from astropy.time.core import Time
from IPython.display import HTML
from numba import njit
from interpolation import interp
import copy

from pyspi.io.get_files import get_files
from pyspi.io.package_data import (get_path_of_internal_data_dir,
                                   get_path_of_external_data_dir)
from pyspi.utils.response.spi_pointing import SPIPointing
from pyspi.utils.response.spi_frame import (transform_icrs_to_spi,
                                            transform_spi_to_icrs)
from pyspi.utils.response.spi_response_data import (ResponseDataPhotopeak,
                                                    ResponseDataRMF)
from pyspi.utils.function_utils import find_needed_ids


@njit
def trapz(y, x):
    """
    Fast trapz integration with numba

    :param x: x values
    :param y: y values

    :returns: Trapz integrated
    """
    return np.trapz(y, x)


@njit
def log_interp1d(x_new, x_old, y_old):
    """
    Linear interpolation in log space for base value pairs (x_old, y_old)
    for new x_values x_new

    :param x_old: Old x values used for interpolation
    :param y_old: Old y values used for interpolation
    :param x_new: New x values

    :returns: y_new from liner interpolation in log space
    """
    # log of all
    logx = np.log10(x_old)
    logxnew = np.log10(x_new)
    
    # Avoid nan entries for yy=0 entries
    logy = np.log10(np.where(y_old <= 0, 1e-99, y_old))
    
    lin_interp = interp(logx, logy, logxnew)

    return np.power(10., lin_interp)


@njit
def add_frac(ph_matrix, i, idx, ebounds, einlow, einhigh):
    """
    Recursive Funktion to get the fraction of einlow...
    """
    if idx+1 == len(ebounds):
        pass
    elif ebounds[idx+1] >= einhigh:
        ph_matrix[i, idx] =\
            np.min(
                np.array(
                    [1,
                     (einhigh-ebounds[idx])/
                     (einhigh-einlow)]
                )
            )

    else:
        frac = np.min(np.array([(ebounds[idx+1]-einlow)/(einhigh-einlow),
                               (ebounds[idx+1]-ebounds[idx])/(einhigh-einlow)]))
        ph_matrix[i, idx] = frac

        add_frac(ph_matrix, i, idx+1, ebounds, einlow, einhigh)


@njit(fastmath=True)
def _get_xy_pos(azimuth, zenith, xmin, ymin, xbin, ybin):
    """
    Get the xy position on the reponse grid for given azimuth and zenith

    :param azimuth: Azmiuth angle in rad
    :param zenith: Zenith angle in rad
    :param xmin: Smallest x-grid entry
    :param ymin: Smallest y-grid entry
    :param xbin: Size of bins in x-direction
    :param ybin: Size of bins in y-direction

    :returns: Grid postition (x,y)
    """

    x = np.cos(azimuth)*np.cos(zenith)
    y = np.sin(azimuth)*np.cos(zenith)
    z = np.sin(zenith)

    zenith_pointing = np.arccos(x)
    azimuth_pointing = np.arctan2(z,y)

    x_pos = (zenith_pointing * np.cos(azimuth_pointing) - xmin) / xbin
    y_pos = (zenith_pointing * np.sin(azimuth_pointing) - ymin) / ybin

    return x_pos, y_pos


def _prep_out_pixels(ix_left, ix_right, iy_low, iy_up):
    """
    Simple function to get the 2D-indices of the 4 points defined by
    given pairs of indices in x and y direction

    :param ix_left: x-index of the left points
    :param ix_right: x-index of the right points
    :param iy_low: y-index of the bottom points
    :param iy_up: y-index of the top points

    :returns: array with the 4 2D indices defining the 4 grid points
    """

    left_low = [int(ix_left), int(iy_low)]
    right_low = [int(ix_right), int(iy_low)]
    left_up = [int(ix_left), int(iy_up)]
    right_up = [int(ix_right), int(iy_up)]

    out = np.array([left_low, right_low, left_up, right_up]).T

    return out


def multi_response_irf_read_objects(times, detector, drm='Photopeak'):
    """
    TODO: This is very ugly. Come up with a better way.
    Function to initalize the needed responses for the given times.
    Only initalize every needed response version once! Because of memory.
    One response object needs about 1 GB of RAM...
    TODO: Not needed at the moment. We need this when we want to analyse
    many pointings together.

    :param times: Times of the different sw used

    :returns: list with correct response version object of the times
    """
    response_versions = []
    for time in times:
        if not time:
            # Default latest response version
            response_versions.append(4)

        elif time < Time(datetime.strptime('031206 060000', '%y%m%d %H%M%S')):
            response_versions.append(0)

        elif time < Time(datetime.strptime('040717 082006', '%y%m%d %H%M%S')):
            response_versions.append(1)

        elif time < Time(datetime.strptime('090219 095957', '%y%m%d %H%M%S')):
            response_versions.append(2)

        elif time < Time(datetime.strptime('100527 124500', '%y%m%d %H%M%S')):
            response_versions.append(3)

        else:
            response_versions.append(4)

    responses = [None, None, None, None, None]

    response_irf_read_times = []
    for version in response_versions:
        if responses[version] is None:
            # Create this response object
            if drm == "Photopeak":
                responses[version] = ResponseDataPhotopeak(
                    detector=detector,
                    version=version)
            else:
                responses[version] = ResponseDataRMF(version=version)
                
        response_irf_read_times.append(responses[version])
    return response_irf_read_times


class ResponseGenerator(object):
    def __init__(self,
                 pointing_id=None,
                 ebounds=None,
                 response_irf_read_object=None,
                 det=None):
        """
        Base Response Class - Here we have everything that stays the same for
        GRB and Constant Pointsource Reponses

        :param ebounds: User defined ebins for binned effective area
        :param response_irf_read_object: Object that holds
            the read in irf values
        :param sc_matrix: Matrix to convert SPI coordinate system <-> ICRS
        :param det: Which detector
        """
        # Get the data, either from afs or from ISDC archive
        try:
            # Get the data from the afs server
            get_files(pointing_id, access="afs")
        except AssertionError:
            # Get the files from the iSDC data archive
            print("AFS data access did not work. "
                  "I will try the ISDC data archive.")
            get_files(pointing_id, access="isdc")

        # Read in geometry file to get sc_matrix
        geometry_file_path = os.path.join(get_path_of_external_data_dir(),
                                          'pointing_data',
                                          pointing_id,
                                          'sc_orbit_param.fits.gz')

        pointing_object = SPIPointing(geometry_file_path)
        sc_matrix = pointing_object.sc_matrix[10]

        self._irf_ob = response_irf_read_object
        self._ebounds = ebounds
        if ebounds is not None:
            self.set_binned_data_energy_bounds(ebounds)
        self._sc_matrix = sc_matrix
        self._det = det
        self._pointing_id = pointing_id

    def set_binned_data_energy_bounds(self, ebounds):
        """
        Change the energy bins for the binned effective_area

        :param ebounds: New ebinedges: ebounds[:-1] start of ebins,
        ebounds[1:] end of ebins

        :returns:
        """

        # if the new bins are not the old ones: update them
        if not np.array_equal(ebounds, self._ebounds):
            self._ene_min = ebounds[:-1]
            self._ene_max = ebounds[1:]
            self._ebounds = ebounds
    
    def get_xy_pos(self, azimuth, zenith):
        """
        Get xy position (in SPI simulation) for given azimuth and zenith

        :param azimuth: Azimuth in Sat. coordinates [rad]
        :param zenith: Zenith in Sat. coordinates [rad]

        :returns: grid position in (x,y) coordinates
        """
        # we call a numba function here to speed it up
        return _get_xy_pos(azimuth,
                           zenith,
                           self.irf_ob.irf_xmin,
                           self.irf_ob.irf_ymin,
                           self.irf_ob.irf_xbin,
                           self.irf_ob.irf_ybin)
    
    def set_location(self, ra, dec):
        """
        Calculate the weighted irfs for the three
        event types for a given position

        :param azimuth: Azimuth position in sat frame
        :param zenith: Zenith position in sat frame

        :returns:
        """

        # Transform ra, dec from icrs to spi frame
        azimuth, zenith = transform_icrs_to_spi(ra,
                                                 dec,
                                                 self._sc_matrix)


        self._weighted_irfs(np.deg2rad(azimuth),
                            np.deg2rad(zenith))

        self._recalculate_response()

    def set_location_direct_sat_coord(self, azimuth, zenith):
        """
        Calculate the weighted irfs for the three
        event types for a given position

        :param azimuth: Azimuth position in sat frame
        :param zenith: Zenith position in sat frame

        :returns: ra and dec value
        """

        self._weighted_irfs(np.deg2rad(azimuth),
                            np.deg2rad(zenith))

        self._recalculate_response()

        return transform_spi_to_icrs(azimuth,
                                     zenith,
                                     self._sc_matrix)

    def _weighted_irfs(self, azimuth, zenith):
        """
        Calculate the weighted irfs for the three event
        types for a given position

        :param azimuth: Azimuth position in sat frame
        :param zenith: Zenith position in sat frame

        :returns:
        """
        raise NotImplementedError("Must be implemented in child class.")

    def _recalculate_response(self):

        raise NotImplementedError("Must be implemented in child class.")

    def _get_irf_weights(self, x_pos, y_pos):
        """
        Get the 4 grid points around (x_pos, y_pos) and the weights
        for a rectangular interpolation

        :param x_pos: grid position x-coordinates
        :param y_pos: grid position y-coordinates

        :returns: weights, x-indices, y-indices of the 4 closest points
        """

        # get the four nearest neighbors
        ix_left = np.floor(x_pos) if (x_pos >= 0.0) else np.floor(x_pos) - 1
        iy_low = np.floor(y_pos) if (y_pos >= 0.0) else np.floor(y_pos) - 1

        ix_right = ix_left + 1
        iy_up = iy_low + 1

        wgt_right = float(x_pos - ix_left)
        wgt_up = float(y_pos - iy_low)

        # pre set the weights
        wgt = np.zeros(4)

        # Get the weights
        if ix_left < 0.:

            if ix_right < 0.:

                out = _prep_out_pixels(ix_left, ix_right, iy_low, iy_up)

                return wgt, out[0], out[1]

            ix_left = ix_right
            wgt_left = 0.5
            wgt_right = 0.5

        elif ix_right >= self.irf_ob.irf_nx:

            if ix_left >= self.irf_ob.irf_nx:

                out = _prep_out_pixels(ix_left, ix_right, iy_low, iy_up)

                return wgt, out[0], out[1]

            ix_right = ix_left
            wgt_left = 0.5
            wgt_right = 0.5

        else:

            wgt_left = 1. - wgt_right

        if iy_low < 0:
            if iy_up < 0:

                out = _prep_out_pixels(ix_left, ix_right, iy_low, iy_up)

                return wgt, out[0], out[1]

            iy_low = iy_up
            wgt_up = 0.5
            wgt_low = 0.5

        elif iy_up >= self.irf_ob.irf_ny:

            if iy_low >= self.irf_ob.irf_ny:

                out = _prep_out_pixels(ix_left, ix_right, iy_low, iy_up)

                return wgt, out[0], out[1]

            iy_up = iy_low
            wgt_up = 0.5
            wgt_low = 0.5

        else:

            wgt_low = 1. - wgt_up

        wgt[0] = wgt_left * wgt_low
        wgt[1] = wgt_right * wgt_low
        wgt[2] = wgt_left * wgt_up
        wgt[3] = wgt_right * wgt_up

        out = _prep_out_pixels(ix_left, ix_right, iy_low, iy_up)

        return wgt, out[0], out[1]

    @property
    def irf_ob(self):
        """
        :returns: the irf_read object with the information from the response
        simulation
        """
        return self._irf_ob

    @property
    def det(self):
        """
        :returns: detector number
        """
        return self._det

    @property
    def ebounds(self):
        """
        :returns: Ebounds of the analysis
        """
        return self._ebounds

    @property
    def ene_min(self):
        """
        :returns: Start of ebounds
        """
        return self._ene_min

    @property
    def ene_max(self):
        """
        :returns: End of Ebounds
        """
        return self._ene_max

    @property
    def rod(self):
        """
        Ensure that you know what you are doing.

        :returns: Roland
        """
        return HTML(filename=os.path.join(
            get_path_of_internal_data_dir(),
            "Roland.html")
                    )


class ResponseRMFGenerator(ResponseGenerator):

    def __init__(self,
                 pointing_id=None,
                 monte_carlo_energies=None,
                 ebounds=None,
                 response_irf_read_object=None,
                 det=None,
                 fixed_rsp_matrix=None):
        """
        Init Response object with total RMF used

        :param pointing_id: The pointing ID for which the
            response should be valid
        :param ebound: Ebounds of Ebins
        :param monte_carlo_energies: Input energy bin edges
        :param response_irf_read_object: Object that holds
            the read in irf values
        :param det: Detector ID
        :param fixed_rsp_matrix: A fixed response matrix to overload
            the normal matrix

        :returns: Object
        """
        assert isinstance(response_irf_read_object, ResponseDataRMF)

        super().__init__(
            pointing_id=pointing_id,
            ebounds=ebounds,
            response_irf_read_object=response_irf_read_object,
            det=det)

        self._monte_carlo_energies = monte_carlo_energies

        if fixed_rsp_matrix is None:
            self._rebin_rmfs()
            self._given_rsp_mat = False
            self._rsp_matrix = None
        else:
            self._given_rsp_mat = True
            self._rsp_matrix = fixed_rsp_matrix

        self._weighted_irf_ph = None
        self._weighted_irf_nonph_1 = None
        self._weighted_irf_nonph_2 = None

    @classmethod
    def from_time(cls,
                  time,
                  det,
                  ebounds,
                  monte_carlo_energies,
                  rsp_read_obj,
                  fixed_rsp_matrix=None):
        """
        Init Response object with total RMF used from a time

        :param time: Time for which to construct the response object
        :param ebound: Ebounds of Ebins
        :param monte_carlo_energies: Input energy bin edges
        :param response_irf_read_object: Object that holds
        the read in irf values
        :param det: Detector ID
        :param fixed_rsp_matrix: A fixed response matrix to overload
        the normal matrix

        :returns: Object
        """

        pointing_id = find_needed_ids(time)

        return cls(
            pointing_id=pointing_id,
            monte_carlo_energies=monte_carlo_energies,
            ebounds=ebounds,
            response_irf_read_object=rsp_read_obj,
            det=det,
            fixed_rsp_matrix=fixed_rsp_matrix
        )

    def _rebin_rmfs(self):
        """
        Rebin the base rmf shape matrices for the given ebounds and
        incoming energies

        :returns:
        """
        # Number of ebins on input and output side
        N_ebins = len(self.ebounds)-1
        N_monte_carlo = len(self.monte_carlo_energies)

        # get the interpolation grid (ein and the log mean of the eout_bins)
        log_eout_mean = 10**((np.log10(self.ebounds[:-1]) +
                              np.log10(self.ebounds[1:])) / 2.0)
        eout_width = self.ebounds[1:] - self.ebounds[:-1]
        xx, yy = np.meshgrid(self.monte_carlo_energies,
                             log_eout_mean)
        points = np.array((xx.ravel(), yy.ravel())).T

        # build the interpolation functions for rmf_2 mat and rmf_3 mat
        base_ebounds = self.irf_ob.ebounds_rmf_2_base
        base_ebounds_width = base_ebounds[1:]-base_ebounds[:-1]

        x_base = self.irf_ob.energies_database
        y_base = 10**((np.log10(base_ebounds[:-1])+
                       np.log10(base_ebounds[1:]))/2.0)

        lin_int_rmf2 =\
            interpolate.RegularGridInterpolator((x_base, y_base),
                                                self.irf_ob.rmf_2_base /
                                                base_ebounds_width,
                                                bounds_error=False,
                                                fill_value=0)

        lin_int_rmf3 =\
            interpolate.RegularGridInterpolator((x_base, y_base),
                                                self.irf_ob.rmf_3_base /
                                                base_ebounds_width,
                                                bounds_error=False,
                                                fill_value=0)

        # call the interpolation and create the correct matrix shape
        mat2inter = (lin_int_rmf2(points).reshape(N_ebins, N_monte_carlo).T)

        mat3inter = (lin_int_rmf3(points).reshape(N_ebins, N_monte_carlo).T)

        # trapz integrate over out bins

        self._mat2inter = 0.5*(mat2inter[1:]+mat2inter[:-1])*eout_width
        self._mat3inter = 0.5*(mat3inter[1:]+mat3inter[:-1])*eout_width

        # normalize the rows

        norm_fact = np.sum(self._mat2inter, axis=1)

        # Normalize this - is this correct?
        self._mat2inter[norm_fact>0] /= norm_fact[norm_fact>0][:, np.newaxis]
        self._mat2inter[norm_fact<=0] = 0

        norm_fact = np.sum(self._mat3inter, axis=1)

        # Normalize this - is this correct?
        self._mat3inter[norm_fact>0] /= norm_fact[norm_fact>0][:, np.newaxis]
        self._mat3inter[norm_fact<=0] = 0

        self._ph_matrix = np.zeros((N_monte_carlo-1,
                                    N_ebins))

        for i, (einlow, einhigh) in \
            enumerate(zip(self.monte_carlo_energies[:-1],
                          self.monte_carlo_energies[1:])):
            if (not einhigh < self.ebounds[0]) and (not einlow > self.ebounds[-1]):
                # find id and fraction
                idx = np.argwhere(self.ebounds > einlow)[0, 0]-1
                if idx==-1:
                    idx=0
                add_frac(self._ph_matrix, i, idx, self.ebounds, einlow, einhigh)

    def _weighted_irfs(self, azimuth, zenith):
        """
        Calculate the weighted irfs for the three event types for a given position

        :param azimuth: Azimuth position in sat frame
        :param zenith: Zenith position in sat frame

        :returns:
        """

        # get the x,y position on the grid
        x, y = self.get_xy_pos(azimuth, zenith)

        # compute the weights between the grids
        wgt, xx, yy = self._get_irf_weights(x, y)


        # If outside of the response pattern set response to zero
        try:
            # select these points on the grid and weight them together
            self._weighted_irf_ph = \
                self.irf_ob.irfs_photopeak[:, self._det, xx, yy].dot(wgt)
            self._weighted_irf_nonph_1 = \
                self.irf_ob.irfs_nonphoto_1[:, self._det, xx, yy].dot(wgt)
            self._weighted_irf_nonph_2 = \
                self.irf_ob.irfs_nonphoto_2[:, self._det, xx, yy].dot(wgt)
        except IndexError:
            self._weighted_irf_ph =\
                np.zeros_like(self.irf_ob.irfs_photopeak[:, self._det, 20, 20])
            self._weighted_irf_nonph_1\
                = np.zeros_like(self.irf_ob.irfs_nonphoto_1[:, self._det, 20, 20])
            self._weighted_irf_nonph_2\
                = np.zeros_like(self.irf_ob.irfs_nonphoto_2[:, self._det, 20, 20])

    def _recalculate_response(self):
        """
        Get response for the current position

        :returns:
        """
        if self._given_rsp_mat:
            self._matrix = self._rsp_matrix
        else:
            ein_bins = np.empty((len(self._monte_carlo_energies)-1, 2))

            ein_bins[:, 0] = self._monte_carlo_energies[:-1]
            ein_bins[:, 1] = self._monte_carlo_energies[1:]

            monte_carlo_log_mean = 10**((
                np.log10(self._monte_carlo_energies[1:]) +
                np.log10(self._monte_carlo_energies[:-1]))/2.0)

            interph = log_interp1d(monte_carlo_log_mean,
                                   self.irf_ob.energies_database,
                                   self._weighted_irf_ph)
            inter2 = log_interp1d(monte_carlo_log_mean,
                                  self.irf_ob.energies_database,
                                  self._weighted_irf_nonph_1)
            inter3 = log_interp1d(monte_carlo_log_mean,
                                  self.irf_ob.energies_database,
                                  self._weighted_irf_nonph_2)

            # TODO clean up these .T calls and maybe change
            # the trapz to logmean

            mat1 = (inter2*self._mat2inter.T).T
            mat2 = (inter3*self._mat3inter.T).T

            # Add photopeak to the ebin with the photon energy within
            # its bounds
            self._transpose_matrix = mat1+mat2+(interph*self._ph_matrix.T).T

            self._matrix = self._transpose_matrix.T

    def clone(self):
        """
        Clone this response object

        :returns: cloned response
        """
        return ResponseRMFGenerator(
            pointing_id=copy.deepcopy(self._pointing_id),
            monte_carlo_energies=copy.deepcopy(self.monte_carlo_energies),
            ebounds=copy.deepcopy(self.ebounds),
            response_irf_read_object=self.irf_ob,
            det=copy.deepcopy(self.det),
            fixed_rsp_matrix=copy.deepcopy(self._rsp_matrix)
        )

    @property
    def matrix(self):
        """
        :returns: response matrix
        """
        return self._matrix

    @property
    def transpose_matrix(self):
        """
        :returns: transposed response matrix
        """
        return self._transpose_matrix

    @property
    def monte_carlo_energies(self):
        """
        :returns: Input energies for response
        """
        return self._monte_carlo_energies


class ResponsePhotopeakGenerator(ResponseGenerator):

    def __init__(self,
                 pointing_id=None,
                 ebounds=None,
                 response_irf_read_object=None,
                 det=None):
        """
        Init Response object with photopeak only

        :param pointing_id: The pointing ID for which the
        response should be valid
        :param ebound: Ebounds of Ebins
        :param response_irf_read_object: Object that holds
        the read in irf values
        :param det: Detector ID
        """
        assert isinstance(response_irf_read_object, ResponseDataPhotopeak)

        # call init of base class
        super(ResponsePhotopeakGenerator, self).__init__(
            pointing_id,
            ebounds,
            response_irf_read_object,
            det)

        self._effective_area = None
        self._weighted_irf_ph = None

    def _weighted_irfs(self, azimuth, zenith):
        """
        Calculate the weighted irfs for the three event types for a given
        position

        :param azimuth: Azimuth position in sat frame
        :param zenith: Zenith position in sat frame

        :returns:
        """

        # get the x,y position on the grid
        x, y = self.get_xy_pos(azimuth, zenith)

        # compute the weights between the grids
        wgt, xx, yy = self._get_irf_weights(x, y)

        # If outside of the response pattern set response to zero
        try:
            # select these points on the grid and weight them together
            self._weighted_irf_ph = \
                self.irf_ob.irfs_photopeak[:,self._det, xx, yy].dot(wgt)

        except IndexError:
            self._weighted_irf_ph =\
                np.zeros_like(self.irf_ob.irfs_photopeak[:, self._det, 20, 20])

    def _recalculate_response(self):
        """
        Get response for a given det

        :param det: Detector ID

        :returns: Full DRM
        """
        #n_energy_bins = len(self._ebounds) - 1
        
        ebins = np.empty((len(self._ene_min), 2))
        eff_area = np.empty_like(ebins)

        ebins[:, 0] = self._ene_min
        ebins[:, 1] = self._ene_max
        
        inter = log_interp1d(self.ebounds,
                             self.irf_ob.energies_database,
                             self._weighted_irf_ph)
        
        eff_area[:, 0] = inter[:-1]
        eff_area[:, 1] = inter[1:]
        
        self._effective_area = trapz(eff_area, ebins)/(self._ene_max -
                                                       self._ene_min)

    @classmethod
    def from_time(cls,
                  time,
                  det,
                  ebounds,
                  rsp_read_obj,):
        """
        Init Response object with photopeak only

        :param time: The time for which the
            response should be valid
        :param ebound: Ebounds of Ebins
        :param response_irf_read_object: Object that holds
            the read in irf values
        :param det: Detector ID

        :return: Object
        """
        pointing_id = find_needed_ids(time)

        return cls(
            pointing_id=pointing_id,
            ebounds=ebounds,
            response_irf_read_object=rsp_read_obj,
            det=det,
        )

    @property
    def effective_area(self):
        """
        :returns: vector with photopeak effective area
        """
        return self._effective_area



from setuptools import setup
import os

import versioneer


# Create list of data files
def find_data_files(directory):

    paths = []

    for (path, directories, filenames) in os.walk(directory):

        for filename in filenames:

            paths.append(os.path.join("..", path, filename))

    return paths


extra_files = find_data_files("pyspi/data")

setup(
    version=versioneer.get_version(),
    include_package_data=True,
    package_data={"": extra_files},
    cmdclass=versioneer.get_cmdclass(),
)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.patches import RegularPolygon
from pyspi.utils.livedets import get_live_dets

NUM_REAL_DETS = 19
NUM_PSEUDO_DOUBLE_DETS = 42
NUM_PSEUDO_TRIPLE_DETS = 42
NUM_TOTAL_DETS = NUM_REAL_DETS + NUM_PSEUDO_DOUBLE_DETS + \
    NUM_PSEUDO_TRIPLE_DETS


# the origins of the detectors
# the underscore keeps these variable from being exposed
# to the user
_detector_origins = ((0, 0), (6, 0), (3, 5.196),
                     (-3, 5.196), (-6, 0), (-3, -5.196),
                     (3, -5.196), (9, -5.196), (12, 0),
                     (9, 5.196), (6, 10.392), (0, 10.392),
                     (-6, 10.392), (-9, 5.196), (-12, 0),
                     (-9, -5.196), (-6, -10.392), (0, -10.392),
                     (6, -10.392))

def _calc_double_origin(det1, det2):
    x = (_detector_origins[det1][0] + _detector_origins[det2][0]) * 0.5
    y = (_detector_origins[det1][1] + _detector_origins[det2][1]) * 0.5

    return x, y

def _construct_double_events_table():
    """

    Helper function to generate double event detector list


    :return:
    """

    # the list of double pairs
    # for details see:
    # https://heasarc.gsfc.nasa.gov/docs/integral/spi/pages/detectors.html
    double_event_pairs = (
        (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (1, 2), (1, 6),
        (1, 7), (1, 8), (1, 9), (2, 3), (2, 9), (2, 10), (2, 11), (3, 4),
        (3, 11), (3, 12), (3, 13), (4, 5), (4, 13), (4, 14), (4, 15), (5, 6),
        (5, 15), (5, 16), (5, 17), (6, 7), (6, 17), (6, 18), (7, 8), (7, 18),
        (8, 9), (9, 10), (10, 11), (11, 12), (12, 13), (13, 14),
        (14, 15), (15, 16), (16, 17), (17, 18))

    # calculate the origins

    origins = np.array([_calc_double_origin(*pair)
                        for pair in double_event_pairs])

    # build a dictionary for out put

    out = dict(detector_number=range(19, 19 + 42),
               x=origins[:, 0],
               y=origins[:, 1],
               detector1=np.array(double_event_pairs)[:, 0],
               detector2=np.array(double_event_pairs)[:, 1]

               )

    # return the as a pandas data frame

    return pd.DataFrame(out)


# build a list that can be exposed to the user
spi_pseudo_double_detectors = _construct_double_events_table()


class SPIDetector(object):

    def __init__(self, detector_number, origin, is_pseudo_detector=False):
        """
        A SPI detector is defined by its number, origin and type
        :param detector_number: the detector number
        :param origin: the detector origin
        :param is_pseudo_detector: if this is a real detector or not
        """
        self._detector_number = detector_number
        self._origin = origin
        self._is_pseudo_detector = is_pseudo_detector
        self._bad = False
        
    @property
    def bad(self):
        return self._bad

    def set_bad(self, flag):
        """
        Set the flag if this is a bad detector
        :param flag: Bad detector?
        :return: 
        """
        self._bad = flag
    
    @property
    def origin(self):
        return self._origin

    @property
    def is_pseudo_detector(self):
        return self._is_pseudo_detector

    @property
    def detector_number(self):
        return self._detector_number


class DoubleEventDetector(SPIDetector):

    def __init__(self, detector_number, origin, detector1, detector2):
        """

        :param detector_number:
        :param origin:
        :param detector1:
        :param detector2:
        """

        super(DoubleEventDetector, self).__init__(detector_number, origin, is_pseudo_detector=True)

        self._detector1 = detector1
        self._detector2 = detector2


# Fill out this class

class TripleEventDetector(SPIDetector):
     pass



class DetectorContents(object):

    def __init__(self, detector_array):
        assert len(detector_array) == NUM_TOTAL_DETS

        self._contents = np.array(detector_array)

        self._real_contents = np.array(detector_array[:NUM_REAL_DETS])

    @classmethod
    def from_spi_data(cls, spi_data):
        pass

    @classmethod
    def from_total_effective_area(cls, spi_response, azimuth, zenith):
        effective_area = spi_response.effective_area_per_detector(azimuth, zenith).sum(axix=0)

        return cls(effective_area)


class SPI(object):

    def __init__(self, bad_detectors=[], time=None):

        self._bad_detectors = bad_detectors
        if time:
            good_detectors = get_live_dets(time)
            self._bad_detectors = [i
                                   for i in range(19)
                                   if i not in good_detectors]
        self._construct_detectors()

    def _construct_detectors(self):
        """

        :return:
        """

        # the real detector origins
        self._detector_origins = _detector_origins
        # go through an build the list of detectors for SPI
        self._detectors = []

        n = 0  # keeps track of the detector number

        for origin in self._detector_origins:

            # first we build the real detectors

            self._detectors.append(SPIDetector(detector_number=n,
                                               origin=origin,
                                               is_pseudo_detector=False))

            n += 1

        # now we build the double event detectors

        for detector in spi_pseudo_double_detectors.iterrows():

            detector = detector[1]

            origin = (detector['x'], detector['y'])

            self._detectors.append(DoubleEventDetector(detector_number=n,
                                                       origin=origin,
                                                       detector1=detector['detector1'],
                                                       detector2=detector['detector2']))

            n += 1

        # TODO: Add the triple event detectors


        for det in self._detectors:
            if det.detector_number in self._bad_detectors:
                det.set_bad(True)

    def plot_spi_working_dets(self,
                              with_pseudo_detectors=True,
                              show_detector_number=True):
        """
        Plot the SPI Detectors and mark the detectors that are not working red
        :param with_pseudo_detectors: Plot pseudo detectors?
        :param show_detector_number: Show the det numbers in the plot?
        :return:
        """
        

        fig, ax = plt.subplots(subplot_kw={'aspect': 'equal'}, figsize=(15,15))

        d=1

        radius = d*2* 1.732

        # first get the colors based of the contents
        #colors, pseudo_colors = self._get_colors_from_contents(cmap,
        #                                                       pseudo_cmap)
        # color iterators
        n = 0
        pseudo_n = 0

        # now we loop over all the detectors and if they have contents
        # we will plot them
        for detector in self._detectors:
            # first the real detectors

            if not detector.is_pseudo_detector:

                if detector.bad:
                    color = "red"
                else:
                    color = "green"

                # create a ploygon and color it based of the contents
                p = RegularPolygon(xy=tuple(d*i for i in detector.origin),
                                   numVertices=6,
                                   radius=radius,
                                   facecolor=color,
                                   ec='k',
                                   lw=3)

                ax.add_patch(p)

                # show the detector number
                if show_detector_number:
                    ax.text(d*detector.origin[0],
                            d*detector.origin[1],
                            detector.detector_number,
                            ha="center",
                            va="center",
                            color='k',
                            size=30)

        ax.set_xlim(d*-16, d*16)
        ax.set_ylim(d*-16, d*16)

        ax.set_yticks([])
        ax.set_xticks([])
        return fig

    def _loadtxt2d(self, intext):
        try:
            return np.loadtxt(intext, ndmin=2)
        except:
            return np.loadtxt(intext)

